from src.spotlight import errors as err
from .validator_test import ValidatorTest


class StopValidationTest(ValidatorTest):
    def test_field_validation_will_stop_if_required(self):
        data = {"test1": "", "test2": "not.a.valid.email"}
        rules = {"test1": "required|email|min:1|max:255", "test2": "required|email"}

        errors = self.validator.validate(data, rules)
        err1 = errors.get("test1")
        err2 = errors.get("test2")

        self.assertEqual(len(errors), 2)
        self.assertEqual(len(err1), 1)
        self.assertEqual(len(err2), 1)
        self.assertEqual(err1[0], err.REQUIRED_ERROR.format(field="test1"))
        self.assertEqual(err2[0], err.EMAIL_ERROR.format(field="test2"))

    def test_nested_list_field_validation_will_stop_if_required(self):
        data = {"test": [{"test1": ""}, {"test1": "not.a.valid.email"}]}
        rules = {"test.*.test1": "required|email|min:1|max:255"}

        errors = self.validator.validate(data, rules)
        err1 = errors.get("test.0.test1")
        err2 = errors.get("test.1.test1")

        self.assertEqual(len(errors), 2)
        self.assertEqual(len(err1), 1)
        self.assertEqual(len(err2), 1)
        self.assertEqual(err1[0], err.REQUIRED_ERROR.format(field="test.0.test1"))
        self.assertEqual(err2[0], err.EMAIL_ERROR.format(field="test.1.test1"))

    def test_nested_dict_field_validation_will_stop_if_required(self):
        data = {"test": {"test1": "", "test2": "not.a.valid.email"}}
        rules = {
            "test.test1": "required|email|min:1|max:255",
            "test.test2": "required|email",
        }

        errors = self.validator.validate(data, rules)
        err1 = errors.get("test.test1")
        err2 = errors.get("test.test2")

        self.assertEqual(len(errors), 2)
        self.assertEqual(len(err1), 1)
        self.assertEqual(len(err2), 1)
        self.assertEqual(err1[0], err.REQUIRED_ERROR.format(field="test.test1"))
        self.assertEqual(err2[0], err.EMAIL_ERROR.format(field="test.test2"))



from datetime import datetime

from src.spotlight import config
from src.spotlight.errors import DATE_TIME_ERROR
from .validator_test import ValidatorTest


class DateTimeTest(ValidatorTest):
    def test_date_time_rule_with_valid_values_expect_no_error(self):
        rules = {"date_time1": "date_time", "date_time2": "date_time"}
        data = {
            "date_time1": "2019-12-31 23:59:59",
            "date_time2": "2019-01-01 00:00:00",
        }
        expected = {}

        errors = self.validator.validate(data, rules)

        self.assertEqual(errors, expected)

    def test_date_time_rule_with_invalid_value_expect_error(self):
        rules = {"date_time1": "date_time", "date_time2": "date_time"}
        data = {"date_time1": "", "date_time2": 102323}

        errors = self.validator.validate(data, rules)

        self.assertEqual(len(errors) > 0, True)
        for field, field_errors in errors.items():
            expected = DATE_TIME_ERROR.format(field=field, format="%Y-%m-%d %H:%M:%S")
            self.assertEqual(field_errors[0], expected)

    def test_valid_date_time_with_valid_values_expect_true(self):
        values = ["2019-12-31 23:59:59", "2019-01-01 00:00:00", "2019-01-01 23:59:59"]

        for value in values:
            actual = self.validator.valid_date_time(value)
            self.assertEqual(actual, True)

    def test_valid_date_time_with_invalid_values_expect_false(self):
        values = [
            "2019-12-01 24:00:00",
            "2019-12-01 00:60:00",
            "2019-12-01 00:00:60",
            "12/32/2019 00:00:00",
            "32/12/2019 00:00:00",
            "12-32-2019 00:00:00",
            "32-12-2019 00:00:00",
            "2019-13-31 00:00:00",
            "2019-12-32 00:00:00",
            "2019-1-1 00:00:00",
            "2019-10-1 00:00:00",
            "2019-1-10 00:00:00",
            "2019-01-01 0:00:00",
            "2019-01-01 00:0:00",
            "2019-01-01 00:00:0",
            True,
            False,
            [],
            {},
            "",
            -1,
            0,
            1,
            12122019,
            "12122019 000000",
        ]

        for value in values:
            actual = self.validator.valid_date_time(value)
            self.assertEqual(actual, False)

    def test_date_time_rule_and_in_rule_with_valid_values_expect_no_error(self):
        rules = {
            "date_time1": "date_time:%Y-%m-%d|in:2019-12-31",
            "date_time2": "date_time|in:2019-01-01 12:00:00",
        }
        data = {"date_time1": "2019-12-31", "date_time2": "2019-01-01 12:00:00"}
        expected = {}

        errors = self.validator.validate(data, rules)

        self.assertEqual(errors, expected)

    def test_date_time_rule_with_format_with_valid_values_expect_no_error(self):
        rules = {"date_time1": "date_time:%Y-%m-%d", "date_time2": "date_time:%m/%d/%Y"}
        data = {"date_time1": "2019-12-31", "date_time2": "01/01/2019"}
        expected = {}

        errors = self.validator.validate(data, rules)

        self.assertEqual(errors, expected)

    def test_date_time_rule_with_format_with_invalid_value_expect_error(self):
        rules = {
            "date_time1": "date_time:%Y-%m-%d",
            "date_time2": "date_time:%Y-%m-%d",
            "date_time3": "date_time:%Y-%m-%d",
        }
        data = {"date_time1": [], "date_time2": {}, "date_time3": 123456}

        errors = self.validator.validate(data, rules)

        for field, field_errors in errors.items():
            expected = DATE_TIME_ERROR.format(field=field, format="%Y-%m-%d")
            self.assertEqual(field_errors[0], expected)

    def test_valid_date_time_with_format_with_valid_values_expect_true(self):
        values = {
            "2019-12-31": "%Y-%m-%d",
            "12/31/2019": "%m/%d/%Y",
            "12312019": "%m%d%Y",
            "2019-01-01 12:00:00": "%Y-%m-%d %H:%M:%S",
            "2019": "%Y",
            "2019, 1, 1": "%Y, %m, %d",
            "Sun": "%a",
            "Thu": "%a",
            "Monday": "%A",
            "Friday": "%A",
            "0": "%w",
            "Jul": "%b",
            "Nov": "%b",
            "February": "%B",
            "August": "%B",
            "31-12-12": "%d-%m-%y",
            "14:24:15": "%H:%M:%S",
            "07:34:42": "%I:%M:%S",
            "am": "%p",
            "pm": "%p",
            "14:24:15:000001": "%H:%M:%S:%f",
            "UTC": "%Z",
            "365": "%j",
            "25": "%U",
            "11": "%W",
            "Tue Aug 16 21:30:00 1988": "%c",
            "08/16/88": "%x",
            "21:30:00": "%X",
            "%": "%%",
            "1": "%u",
            "1985-08-23T3:00:00.000": "%Y-%m-%dT%H:%M:%S.%f",
            "2019-05-17T19:59:19+0000": "%Y-%m-%dT%H:%M:%S%z",
        }

        for value, date_format in values.items():
            actual = self.validator.valid_date_time(value, date_format)
            self.assertEqual(actual, True)

    def test_date_time_rule_with_date_time_objects_expect_no_errors(self):
        dt1 = datetime.strptime("2019-05-16T20:59:19+0000", "%Y-%m-%dT%H:%M:%S%z")
        dt2 = datetime.strptime("2019-05-17 19:59:19", config.DEFAULT_DATE_TIME_FORMAT)
        dt3 = datetime.strptime("22/09/2020 18:59:19+0400", "%d/%m/%Y %H:%M:%S%z")

        rules = {
            "date_time1": "date_time",
            "date_time2": "date_time",
            "date_time3": "date_time",
        }
        data = {"date_time1": dt1, "date_time2": dt2, "date_time3": dt3}
        expected = {}

        errors = self.validator.validate(data, rules)

        self.assertEqual(errors, expected)

    def test_valid_date_time_with_format_with_date_time_object_expect_true(self):
        dt = datetime.strptime("2019-05-17 19:59:19", config.DEFAULT_DATE_TIME_FORMAT)

        actual = self.validator.valid_date_time(dt)
        self.assertEqual(actual, True)

    def test_valid_date_time_with_format_with_invalid_values_expect_false(self):
        values = [
            "12/32/2019",
            "32/12/2019",
            "12-32-2019",
            "32-12-2019",
            "2019-13-31",
            "2019-12-32",
            True,
            False,
            [],
            {},
            "",
            -1,
            0,
            1,
            12122019,
            "12122019",
        ]
        date_time_formats = [
            "%m/%d/%Y",
            "%d/%m/%Y",
            "%m-%d-%Y",
            "%d-%m-%Y",
            "%Y-%m-%d",
            "%Y-%m-%d",
            "%Y-%m-%d",
            "%Y-%m-%d",
            "%Y-%m-%d",
            "%Y-%m-%d",
            "%Y-%m-%d",
            "%Y-%m-%d",
            "%Y-%m-%d",
            "%Y-%m-%d",
            "%Y-%m-%d",
            "%Y-%m-%d",
        ]

        for value, date_time_format in zip(values, date_time_formats):
            actual = self.validator.valid_date_time(value, date_time_format)
            self.assertEqual(actual, False)



from src.spotlight.errors import FILLED_ERROR
from .validator_test import ValidatorTest


class FilledTest(ValidatorTest):
    def setUp(self):
        self.field = "test"
        self.filled_error = FILLED_ERROR.format(field=self.field)
        self.rules = {"test": "filled"}

    def test_filled_rule_with_empty_string_expect_error(self):
        data = {"test": ""}
        expected = self.filled_error

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_filled_rule_with_field_present_expect_no_error(self):
        data = {"test": "hello"}
        expected = None

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)

    def test_filled_rule_with_boolean_false_expect_no_error(self):
        data = {"test": False}
        expected = None

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)

    def test_filled_rule_with_spaces_expect_error(self):
        data = {"test": "   "}
        expected = self.filled_error

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_filled_rule_with_none_expect_error(self):
        data = {"test": None}
        expected = self.filled_error

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_filled_rule_with_no_input_expect_no_error(self):
        data = {}
        expected = None

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)



from src.spotlight.errors import MIN_STRING_ERROR, IN_ERROR, REQUIRED_IF_ERROR
from .validator_test import ValidatorTest


class CustomMessageTest(ValidatorTest):
    def setUp(self):
        self.validator.overwrite_messages = {}
        self.validator.overwrite_fields = {}
        self.validator.overwrite_values = {}

    def test_custom_rule_message_expect_new_message(self):
        rules = {"some_field": "email"}
        data = {"some_field": "this.is.not.a.valid.email"}
        new_message = "You've supplied an invalid e-mail address."
        self.validator.overwrite_messages = {"email": new_message}

        errors = self.validator.validate(data, rules)
        errs = errors.get("some_field")

        self.assertEqual(errs[0], new_message)

    def test_custom_field_rule_message_expect_new_message(self):
        rules = {"some_field": "required"}
        data = {}
        new_message = "Hey! This is a required field!"
        self.validator.overwrite_messages = {"some_field.required": new_message}

        errors = self.validator.validate(data, rules)
        errs = errors.get("some_field")

        self.assertEqual(errs[0], new_message)

    def test_custom_field_rule_message_with_field_param_expect_new_message(self):
        field = "some_field"
        rules = {"some_field": "required"}
        data = {}
        new_message = "Hey! The {field} field is a required field!"
        self.validator.overwrite_messages = {"some_field.required": new_message}

        errors = self.validator.validate(data, rules)
        errs = errors.get(field)

        self.assertEqual(errs[0], new_message.format(field=field))

    def test_custom_field_rule_message_with_min_param_expect_new_message(self):
        field = "some_field"
        rules = {"some_field": "min:5"}
        data = {"some_field": "oops"}
        new_message = "Hey! The {field} field has to be at least {min} chars!"
        self.validator.overwrite_messages = {"some_field.min": new_message}

        errors = self.validator.validate(data, rules)
        errs = errors.get(field)

        self.assertEqual(errs[0], new_message.format(field=field, min=5))

    def test_custom_field_expect_new_field(self):
        rules = {"some_field": "min:5"}
        data = {"some_field": "oops"}
        self.validator.overwrite_fields = {"some_field": "custom"}
        expected = MIN_STRING_ERROR.format(field="custom", min=5)

        errors = self.validator.validate(data, rules)
        errs = errors.get("some_field")

        self.assertEqual(errs[0], expected)

    def test_custom_field_and_custom_other_field_expect_new_fields(self):
        rules = {"some_field1": "required", "some_field2": "required_with:some_field1"}
        data = {"some_field1": "hello"}
        new_message = "The {field} field has to be present with {other}."
        self.validator.overwrite_messages = {"some_field2.required_with": new_message}
        self.validator.overwrite_fields = {
            "some_field1": "custom1",
            "some_field2": "custom2",
        }
        expected = new_message.format(field="custom2", other="custom1")

        errors = self.validator.validate(data, rules)
        errs = errors.get("some_field2")

        self.assertEqual(errs[0], expected)

    def test_custom_field_message_with_custom_field_expect_new_message(self):
        rules = {"email": "email"}
        data = {"email": "this.is.not.a.valid.email"}
        new_message = "You've supplied an invalid {field}."
        self.validator.overwrite_messages = {"email": new_message}
        self.validator.overwrite_fields = {"email": "e-mail address"}
        expected = new_message.format(field="e-mail address")

        errors = self.validator.validate(data, rules)
        errs = errors.get("email")

        self.assertEqual(errs[0], expected)

    def test_custom_field_values_expect_errors_with_new_values(self):
        field = "some_field"
        new_values = "new1, new2, new3"
        rules = {"some_field": "in:val1,val2,val3"}
        data = {"some_field": "this.is.not.a.valid.email"}
        self.validator.overwrite_values = {
            "some_field": {"val1": "new1", "val2": "new2", "val3": "new3"}
        }
        expected = IN_ERROR.format(field=field, values=new_values)

        errors = self.validator.validate(data, rules)
        errs = errors.get(field)

        self.assertEqual(errs[0], expected)

    def test_custom_global_values_expect_errors_with_new_values(self):
        rules = {"field1": "in:crypto,cc,ideal", "field2": "in:crypto,cc,ideal"}
        data = {"field1": "test", "field2": "test"}
        self.validator.overwrite_values = {"cc": "credit card"}
        values = "crypto, credit card, ideal"

        errors = self.validator.validate(data, rules, flat=True)

        self.assertEqual(len(errors), 2)
        self.assertEqual(errors[0], IN_ERROR.format(field="field1", values=values))
        self.assertEqual(errors[1], IN_ERROR.format(field="field2", values=values))

    def test_custom_field_value_expect_error_with_new_value(self):
        rules = {"test": "required_if:other_field,cc"}
        data = {"other_field": "cc"}
        self.validator.overwrite_values = {"test": {"cc": "credit card"}}
        expected = REQUIRED_IF_ERROR.format(
            field="test", other="other_field", value="credit card"
        )

        errors = self.validator.validate(data, rules, flat=True)
        actual = errors[0]

        self.assertEqual(actual, expected)

    def test_custom_other_field_value_expect_error_with_new_value(self):
        rules = {
            "payment_type": "in:crypto,cc,ideal",
            "credit_card_number": "required_if:payment_type,cc",
        }
        data = {"payment_type": "cc"}
        self.validator.overwrite_values = {"payment_type": {"cc": "credit card"}}
        expected = REQUIRED_IF_ERROR.format(
            field="credit_card_number", other="payment_type", value="credit card"
        )

        errors = self.validator.validate(data, rules, flat=True)
        actual = errors[0]

        self.assertEqual(actual, expected)

    def test_custom_message_and_fields_and_value_expect_custom_error(self):
        new_message = "The {field} field is required when the {other} is {value}."
        field = "credit card number"
        other = "payment type"
        value = "credit card"

        rules = {
            "payment_type": "in:crypto,cc,ideal",
            "credit_card_number": "required_if:payment_type,cc",
        }
        data = {"payment_type": "cc"}

        self.validator.overwrite_messages = {
            "credit_card_number.required_if": new_message
        }
        self.validator.overwrite_fields = {
            "payment_type": other,
            "credit_card_number": field,
        }
        self.validator.overwrite_values = {"cc": value}
        expected = new_message.format(field=field, other=other, value=value)

        errors = self.validator.validate(data, rules)
        errs = errors.get("credit_card_number")

        self.assertEqual(expected, errs[0])



from typing import Pattern, AnyStr, Any

from . import config
from .exceptions import FieldValueNotFoundError


def regex_match(compiled_regex: Pattern[AnyStr], value: Any) -> bool:
    """Checks if the value is a full match of the compiled regex"""
    try:
        match = compiled_regex.fullmatch(value)
    except:
        return False
    else:
        return match is not None


def equal(*values: Any) -> bool:
    """Checks if passed values are equal"""
    return len(set([str(v) for v in values])) == 1


def missing(data, field) -> bool:
    """Checks if the field is missing from data"""
    try:
        _get_field_value(data, field)
    except FieldValueNotFoundError:
        return True

    return False


def get_field_value(data, field) -> Any:
    """Return the value of the field in the data"""
    try:
        return _get_field_value(data, field)
    except FieldValueNotFoundError:
        return None


def _get_field_value(value, field):
    segments = field.split(config.FIELD_DELIMITER)
    try:
        for key in segments:
            if not isinstance(value, dict) and not isinstance(value, list):
                value = value.__dict__
            if key.isnumeric():
                value = value[int(key)]
            else:
                value = value[key]
    except (TypeError, AttributeError, KeyError, IndexError):
        raise FieldValueNotFoundError

    return value


def empty(value) -> bool:
    """Checks if the value is empty"""

    # Value is None
    if value is None:
        return True

    # Empty string
    if isinstance(value, str):
        if value.strip() == "":
            return True

    # Empty list or empty dict
    if isinstance(value, list) or isinstance(value, dict):
        if len(value) == 0:
            return True

    return False



from src.spotlight.errors import IN_ERROR
from .validator_test import ValidatorTest


class InTest(ValidatorTest):
    def setUp(self):
        self.field = "test"
        self.in_error = IN_ERROR.format(field=self.field, values="val0, val1, val2")

    def test_in_rule_with_invalid_value_expect_error(self):
        data = {"test": "val3"}
        rules = {"test": "in:val0,val1,val2"}
        expected = self.in_error

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_in_rule_with_valid_value_expect_no_error(self):
        data = {"test": "val1"}
        rules = {"test": "in:val0,val1,val2"}
        expected = None

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)

    def test_in_rule_with_various_string_capitalization_expect_no_error(self):
        data = {
            "case1": "camelCase",
            "case2": "snake_case",
            "case3": "kebab-case",
            "case4": "PascalCase",
            "case5": "lowercase",
            "case6": "UPPERCASE",
            "case7": "AlTcAsE",
        }
        valid_values = (
            "camelCase,snake_case,kebab-case,PascalCase,lowercase,UPPERCASE,AlTcAsE"
        )
        rules = {
            "case1": f"in:{valid_values}",
            "case2": f"in:{valid_values}",
            "case3": f"in:{valid_values}",
            "case4": f"in:{valid_values}",
            "case5": f"in:{valid_values}",
            "case6": f"in:{valid_values}",
        }

        errors = self.validator.validate(data, rules)

        self.assertEqual(len(errors.items()), 0)

    def test_in_rule_with_mismatching_capitalization_expect_error(self):
        data = {"test1": "VAL1", "test2": "vAl1", "test3": "VaL1"}
        rules = {
            "test1": "in:val0,val1,val2",
            "test2": "in:val0,VaL1,val2",
            "test3": "in:val0,vAl1,val2",
        }

        errors = self.validator.validate(data, rules)

        self.assertEqual(len(errors), 3)

    def test_in_rule_with_integer_expect_no_error(self):
        data = {"test": 1}
        rules = {"test": "in:1,2,3"}
        expected = None

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)

    def test_in_rule_with_bool_expect_no_error(self):
        data = {"test": True}
        rules = {"test": "in:True,False"}
        expected = None

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)



import unittest
from unittest import mock

from src.spotlight.validator import Validator


class ValidatorTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.validator = Validator()


class ValidatorPluginTest(unittest.TestCase):
    def test_plugin_rules_registered(self):
        plugin = Validator.Plugin()
        with mock.patch.object(plugin, "rules", wraps=plugin.rules) as m:
            Validator([plugin])
        m.assert_called()


class DirectValidationMethodsTest(ValidatorTest):
    def test_direct_validation_methods_are_present(self):
        methods = [attr for attr in dir(Validator) if attr.startswith("valid_")]

        self.assertEqual(len(methods) > 0, True)
        for name, rule in self.validator._available_rules.items():
            for attr in dir(rule):
                if attr.startswith("valid_"):
                    self.assertEqual(attr in methods, True)



class RuleNotFoundError(Exception):
    def __init__(self, rule: str):
        super().__init__(f"the '{rule}' rule does not exist")


class RuleNameAlreadyExistsError(Exception):
    def __init__(self, rule: str):
        super().__init__(f"the rule name '{rule}' already exists")


class InvalidDataError(Exception):
    def __init__(self, data_type: type):
        super().__init__(
            f"expected 'dict' or an 'object' that can be converted to a 'dict' got '{data_type}'"
        )


class InvalidRulesError(Exception):
    def __init__(self, rules_type: type):
        super().__init__(f"expected 'dict' got '{rules_type}'")


class AttributeNotImplementedError(Exception):
    def __init__(self, attribute: str, class_name: str):
        super().__init__(
            f"the class '{class_name}' is missing the '{attribute}' attribute"
        )


class FieldValueNotFoundError(Exception):
    pass


class InvalidRuleParameters(Exception):
    def __init__(self):
        super().__init__(
            f"something went wrong parsing the rule parameters, make sure the format is correct"
        )


class InvalidDateTimeFormat(Exception):
    def __init__(self):
        super().__init__(f"invalid date time format")



from decimal import Decimal

from src.spotlight.errors import DECIMAL_ERROR
from .validator_test import ValidatorTest


class DecimalTest(ValidatorTest):
    def setUp(self):
        self.field = "test"
        self.decimal_error = DECIMAL_ERROR.format(field=self.field)

    def test_decimal_rule_with_string_expect_error(self):
        rules = {"test": "decimal"}
        data = {"test": "test"}
        expected = self.decimal_error

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_decimal_rule_with_integer_expect_error(self):
        rules = {"test": "decimal"}
        data = {"test": 1}
        expected = self.decimal_error

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_decimal_rule_with_float_expect_error(self):
        rules = {"test": "decimal"}
        data = {"test": 1.0}
        expected = self.decimal_error

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_decimal_rule_with_ten_expect_no_error(self):
        rules = {"test": "decimal"}
        data = {"test": Decimal("10.0")}
        expected = None

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)

    def test_valid_decimals_with_invalid_values_expect_false(self):
        invalid_decimals = [0, 1, "0", "1", -1.0, 1.4, [], {}, True, False, "test"]

        for invalid_decimal in invalid_decimals:
            actual = self.validator.valid_decimal(invalid_decimal)
            self.assertEqual(actual, False)

    def test_valid_decimals_with_valid_values_expect_true(self):
        valid_decimals = [Decimal("-1.2"), Decimal("1.2"), Decimal("1"), Decimal("100")]

        for valid_decimal in valid_decimals:
            actual = self.validator.valid_decimal(valid_decimal)
            self.assertEqual(actual, True)



from typing import Any, List

import pytest

from src.spotlight.exceptions import (
    AttributeNotImplementedError,
    RuleNameAlreadyExistsError,
    InvalidDataError,
    InvalidRulesError,
    RuleNotFoundError,
)
from src.spotlight.rules import Rule
from .validator_test import ValidatorTest


class ExceptionsTest(ValidatorTest):
    def test_new_rule_without_name_expect_error(self):
        with pytest.raises(AttributeNotImplementedError):

            class NewRule(Rule):
                def passes(
                    self, field: str, value: Any, parameters: List[str], validator
                ) -> bool:
                    return True

                @property
                def message(self) -> str:
                    return "Hello World!"

            self.validator.register_rule(NewRule())

    def test_new_rule_with_existing_name_expect_error(self):
        with pytest.raises(RuleNameAlreadyExistsError):

            class NewRule(Rule):
                name = "required"

                def passes(
                    self, field: str, value: Any, parameters: List[str], validator
                ) -> bool:
                    return True

                @property
                def message(self) -> str:
                    return "Hello World!"

            self.validator.register_rule(NewRule())

    def test_validate_with_invalid_data_type_expect_error(self):
        with pytest.raises(InvalidDataError):
            self.validator.validate(data=[], rules={})

    def test_validate_with_invalid_rules_type_expect_error(self):
        with pytest.raises(InvalidRulesError):
            self.validator.validate(data={}, rules=[])

    def test_empty_rule_expect_error(self):
        rules = {"test": ""}
        data = {"test": "John Doe"}

        with pytest.raises(RuleNotFoundError):
            self.validator.validate(data, rules)

    def test_non_existent_rule_expect_rule_not_found_error(self):
        rules = {"test": "not_a_rule"}
        data = {"test": "John Doe"}

        with pytest.raises(RuleNotFoundError):
            self.validator.validate(data, rules)

    def test_rule_abstract_methods_with_super_call_expect_not_implemented_error(self):
        class NewRule(Rule):
            name = "new_rule"

            def passes(
                self, field: str, value: Any, parameters: List[str], validator
            ) -> bool:
                return super().passes(field, value, parameters, validator)

            @property
            def message(self) -> str:
                return super().message

        with pytest.raises(NotImplementedError):
            msg = NewRule().message

        with pytest.raises(NotImplementedError):
            NewRule().passes("", None, [], {})



from datetime import datetime, timedelta

import pytest

from src.spotlight.errors import AFTER_ERROR
from src.spotlight.exceptions import InvalidDateTimeFormat
from .validator_test import ValidatorTest


class AfterTest(ValidatorTest):
    def test_after_rule_with_valid_value_expect_no_error(self):
        rules = {"end_time": "date_time|after:2019-08-24 16:28:00"}
        data = {"end_time": "2019-08-24 16:48:00"}
        expected = {}

        errors = self.validator.validate(data, rules)

        self.assertEqual(errors, expected)

    def test_after_rule_with_value_and_invalid_time_expect_error(self):
        field = "end_time"
        rules = {"end_time": "date_time|after:2019-08-24 17:00:00"}
        data = {"end_time": "2019-08-24 16:28:00"}
        expected = AFTER_ERROR.format(field=field, other="2019-08-24 17:00:00")

        errors = self.validator.validate(data, rules)
        errs = errors.get(field)

        self.assertEqual(errs[0], expected)

    def test_after_rule_with_invalid_values_expect_exception(self):
        values = [
            "after",
            "after:",
            "after:2019-08-24",
            "after:invalid",
            "after:2019-24-08 16:27:00",
            "after:dt2",
        ]
        data = {"dt1": "2019-08-24 16:28:00", "dt2": "invalid"}

        for value in values:
            with pytest.raises(InvalidDateTimeFormat):
                rules = {"dt1": f"date_time|{value}"}
                self.validator.validate(data, rules)

    def test_after_rule_with_custom_format_and_valid_value_expect_no_error(self):
        rules = {"end_time": "date_time:%m/%d/%Y %H:%M:%S|after:2019-08-24 16:46:00"}
        data = {"end_time": "08/24/2019 16:47:00"}
        expected = {}

        errors = self.validator.validate(data, rules)

        self.assertEqual(errors, expected)

    def test_after_rule_with_custom_format_and_invalid_value_expect_error(self):
        field = "end_time"
        rules = {"end_time": "date_time:%m/%d/%Y %H:%M:%S|after:2019-08-24 16:48:00"}
        data = {"end_time": "08/24/2019 16:48:00"}
        expected = AFTER_ERROR.format(field=field, other="2019-08-24 16:48:00")

        errors = self.validator.validate(data, rules)
        errs = errors.get(field)

        self.assertEqual(errs[0], expected)

    def test_after_rule_with_valid_other_field_expect_no_error(self):
        rules = {"start_time": "date_time", "end_time": "date_time|after:start_time"}
        data = {"start_time": "2019-08-24 16:28:00", "end_time": "2019-08-24 16:48:00"}
        expected = {}

        errors = self.validator.validate(data, rules)

        self.assertEqual(errors, expected)

    def test_after_rule_with_valid_other_field_expect_error(self):
        field = "end_time"
        rules = {"start_time": "date_time", "end_time": "date_time|after:start_time"}
        data = {"start_time": "2019-08-24 16:48:00", "end_time": "2019-08-24 16:48:00"}
        expected = AFTER_ERROR.format(field=field, other="start_time")

        errors = self.validator.validate(data, rules)
        errs = errors.get(field)

        self.assertEqual(errs[0], expected)

    def test_after_rule_with_invalid_other_field_expect_exception(self):
        rules = {"start_time": "date_time", "end_time": "date_time|after:start_time"}
        data = {"start_time": "invalid", "end_time": "2019-08-24 16:48:00"}

        with pytest.raises(InvalidDateTimeFormat):
            self.validator.validate(data, rules)

    def test_after_rule_with_format_value_expect_no_error(self):
        rules = {"start_time": "date_time:%H:%M:%S|after:12:00:00"}
        data = {"start_time": "12:00:01"}
        expected = {}

        errors = self.validator.validate(data, rules)

        self.assertEqual(errors, expected)

    def test_after_rule_with_invalid_value_expect_exception(self):
        rules = {"start_time": "date_time|after:12:00:00"}
        data = {"start_time": "2019-06-01 12:00:01"}

        with pytest.raises(InvalidDateTimeFormat):
            self.validator.validate(data, rules)

    def test_after_rule_with_python_datetime_objects_value_expect_no_error(self):
        time = datetime.utcnow()
        rules = {"start_time": "date_time", "end_time": "date_time|after:start_time"}
        data = {"start_time": time, "end_time": time + timedelta(days=1)}
        expected = {}

        errors = self.validator.validate(data, rules)

        self.assertEqual(errors, expected)

    def test_after_rule_with_python_datetime_object_and_invalid_time_value_expect_error(self):
        time = datetime.utcnow()
        field = "end_time"
        rules = {"start_time": "date_time", "end_time": "date_time|after:start_time"}
        data = {"start_time": time, "end_time": time}
        expected = AFTER_ERROR.format(field=field, other="start_time")

        errors = self.validator.validate(data, rules)
        errs = errors.get(field)

        self.assertEqual(errs[0], expected)



from src.spotlight.errors import REQUIRED_UNLESS_ERROR
from .validator_test import ValidatorTest


class RequiredUnlessTest(ValidatorTest):
    def setUp(self):
        self.other_field = "test1"
        self.field = "test2"
        self.value = "test"
        self.required_unless_error = REQUIRED_UNLESS_ERROR.format(
            field=self.field, other=self.other_field, value=self.value
        )

    def test_required_unless_rule_with_required_condition_expect_error(self):
        data = {"test1": "test2"}
        rules = {"test2": "required_unless:test1,test"}
        expected = self.required_unless_error

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_required_unless_rule_with_not_required_condition_expect_no_error(self):
        data = {"test1": "test"}
        rules = {"test2": "required_unless:test1,test"}
        expected = None

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)

    def test_required_unless_rule_with_missing_field_expect_error(self):
        data = {}
        rules = {"test2": "required_unless:test1,test"}
        expected = self.required_unless_error

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)



import re
from string import Formatter
from typing import Union, List, overload, Tuple, Iterator, Dict, Any

from . import rules as rls, config
from .exceptions import RuleNotFoundError, InvalidDataError, InvalidRulesError
from .utils import get_field_value


class Validator:
    """
    Creates an instance of the Validator class.

    Parameters
    ----------
    plugins : list
        A list of plugins that add additional validation rules.
    """

    class Plugin:
        """
        Creates an instance of the Validator Plugin class.
        """

        def rules(self) -> List[rls.Rule]:
            return []

    def __init__(self, plugins: List[Plugin] = None):
        self.data = None
        self.rules = None
        self.output = {}
        self.config = config
        self._flat_list = []

        self.overwrite_messages = {}
        self.overwrite_fields = {}
        self.overwrite_values = {}

        self._implicit_rules: List[str] = []
        self._available_rules: Dict[str, rls.Rule] = {}

        self._setup_default_rules()
        self._setup_plugins(plugins or [])

    def _setup_default_rules(self):
        self.register_rules(self._default_rules())

    def register_rules(self, rules: [rls.Rule]):
        for rule in rules:
            self.register_rule(rule)

    def register_rule(self, rule: rls.Rule):
        self._setup_rule(rule)

    def _setup_rule(self, rule):
        self._available_rules[rule.name] = rule
        if rule.implicit:
            self._implicit_rules.append(rule.name)

    @staticmethod
    def _default_rules() -> List[rls.Rule]:
        return [
            rls.AcceptedRule(),
            rls.AfterRule(),
            rls.AlphaNumRule(),
            rls.AlphaNumSpaceRule(),
            rls.BeforeRule(),
            rls.BooleanRule(),
            rls.DateTimeRule(),
            rls.DecimalRule(),
            rls.DictRule(),
            rls.EmailRule(),
            rls.EndsWithRule(),
            rls.FilledRule(),
            rls.FloatRule(),
            rls.InRule(),
            rls.IntegerRule(),
            rls.IpRule(),
            rls.JsonRule(),
            rls.ListRule(),
            rls.MaxRule(),
            rls.MinRule(),
            rls.NotWithRule(),
            rls.RequiredIfRule(),
            rls.RequiredRule(),
            rls.RequiredUnlessRule(),
            rls.RequiredWithRule(),
            rls.RequiredWithoutRule(),
            rls.SizeRule(),
            rls.StartsWithRule(),
            rls.StringRule(),
            rls.UrlRule(),
            rls.Uuid4Rule(),
        ]

    def _setup_plugins(self, plugins: List[Plugin]):
        for plugin in plugins:
            self.register_rules(plugin.rules())

    @staticmethod
    def _dynamically_add_static_validation_methods(rule):
        for attr in dir(rule):
            if attr.startswith("valid_"):
                validation_method = getattr(rule, attr)
                setattr(Validator, attr, staticmethod(validation_method))

    @overload
    def validate(
        self, data: dict, rules: dict, flat: bool = False
    ) -> Union[dict, list]:
        ...

    @overload
    def validate(
        self, data: object, rules: dict, flat: bool = False
    ) -> Union[dict, list]:
        ...

    def validate(
        self, data: Union[dict, object], rules: dict, flat: bool = False
    ) -> Union[dict, list]:
        """
        Validate data with given rules.

        Parameters
        ----------
        data : dict or object
            Dict or object that can be converted to a dict with data that needs
            to be validated.
            For example: {"email": "john.doe@example.com"},
            CustomDataClass(email="john.doe@example.com")
        rules : dict
            Dict with validation rules for the given data.
            For example: {"email": "required|email|unique:user,email"}
        flat : bool, optional
            Returns a list of errors instead of a dict if true.

        Returns
        -------
        errors : dict or list
            Dict or list of errors. When a dict of errors is returned, each key
            represents the field name, and the corresponding value is a list of
            errors for that key/field. When the optional parameter `flat` is
            set to true, a list of only the error messages is returned.
        """
        self.data = data
        self.rules = rules
        self.output = {}

        self._convert_data_to_dict()
        self._validate_rules_type()
        self._validate_data()

        if flat:
            self._flat_list = []
            self._flatten_output(self.output)

            return self._flat_list

        return self.output

    def _sub_fields(self, field) -> Iterator[str]:
        if not self._contains_wildcard(field):
            yield field
            return

        root = field.split(self.config.FIELD_WILD_CARD)[0].strip(
            self.config.FIELD_DELIMITER
        )
        root_value = self._get_field_value(root)

        if isinstance(root_value, list):
            for i, _ in enumerate(root_value):
                new_field = field.replace(self.config.FIELD_WILD_CARD, str(i), 1)
                yield from self._sub_fields(new_field)

    def _validate_data(self):
        # Iterate over fields
        for raw_field, rules in self._field_iterator():
            # Iterate over sub fields
            for field in self._sub_fields(raw_field):
                # Iterate over rules
                for rule_name, rule_parameters in self.rule_iterator(rules):
                    # Check if rule exists
                    if not self._rule_exists(rule_name):
                        raise RuleNotFoundError(rule_name)

                    # Check if field is validatable
                    if self._is_validatable(field, rule_name):
                        rule = self._available_rules.get(rule_name)
                        value = self._get_field_value(field)

                        # If rule didn't pass, add error
                        if not rule.passes(field, value, rule_parameters, self):
                            self._add_error(rule)

                            # Stop
                            if rule.stop:
                                break

    def _field_iterator(self) -> Iterator[Tuple[str, List[str]]]:
        for field, rules in self.rules.items():
            yield field, self._split_rules(rules)

    def rule_iterator(self, rules) -> Iterator[Tuple[str, List[str]]]:
        for rule in rules:
            yield self._rule_name(rule), self._rule_parameters(rule)

    def _split_rules(self, rules: str) -> List[str]:
        return rules.split(self.config.RULE_DELIMITER)

    def _rule_name(self, rule: str) -> str:
        return self._split_rule(rule)[0]

    def _rule_parameters(self, rule: str) -> List[str]:
        if self.config.RULE_PARAM_DELIMITER not in rule:
            return []

        return self._split_rule(rule)[1].split(self.config.RULE_PARAMS_DELIMITER)

    def _split_rule(self, rule: str) -> List[str]:
        return rule.split(self.config.RULE_PARAM_DELIMITER, 1)

    def _convert_data_to_dict(self):
        if not isinstance(self.data, dict):
            try:
                self.data = self.data.__dict__
            except AttributeError:
                raise InvalidDataError(type(self.data))

    def _validate_rules_type(self):
        if type(self.rules) is not dict:
            raise InvalidRulesError(type(self.rules))

    def _rule_exists(self, rule_name: str) -> bool:
        return rule_name in self._available_rules

    def _contains_wildcard(self, value) -> bool:
        return self.config.FIELD_WILD_CARD in value

    def _is_validatable(self, field: str, rule_name: str) -> bool:
        return self._field_is_present(field) or self._rule_is_implicit(rule_name)

    def _field_is_present(self, field: str) -> bool:
        return self._get_field_value(field) is not None

    def _rule_is_implicit(self, rule_name: str) -> bool:
        return rule_name in self._implicit_rules

    def _add_error(self, rule: rls.Rule):
        field = rule.message_fields.get(self.config.FIELD_KEY)
        error = self._create_error(rule)

        if field in self.output:
            self.output.get(field).append(error)
        else:
            self.output[field] = [error]

    @staticmethod
    def _find_fields_in_error(error: str, regex: bool = False) -> list:
        if regex:
            return re.findall("{(.*?)}", error)

        return [x[1] for x in Formatter().parse(error) if x[1] is not None]

    def _create_error(self, rule: rls.Rule):
        error = rule.message
        fields = rule.message_fields
        field = fields.get(self.config.FIELD_KEY)
        field = self._convert_field_to_wildcard_field(field)
        combined_field = field + self.config.FIELD_DELIMITER + rule.name

        # Overwrite values
        fields = self._overwrite_values(field, fields)

        # If present, format values
        if "values" in fields:
            fields["values"] = self._format_values(fields["values"])

        # Overwrite messages
        error = self._overwrite_messages(error, rule.name, field, combined_field)

        # Overwrite fields
        fields = self._overwrite_fields(fields)

        return error.format(**fields)

    def _overwrite_values(self, field, fields) -> dict:
        for key in ["value", "values"]:
            new_values = fields.get(key)
            field_overwrite_values = self.overwrite_values.get(field, [])

            # Values
            if isinstance(new_values, list):
                # Field
                self._overwrite_values_in_list(
                    fields, new_values, field_overwrite_values, key
                )
                # Global
                self._overwrite_values_in_list(
                    fields, new_values, self.overwrite_values, key
                )
            # Value
            elif isinstance(new_values, str):
                # Field
                self._overwrite_value(fields, new_values, field_overwrite_values, key)
                # Global
                self._overwrite_value(fields, new_values, self.overwrite_values, key)

            # Other field
            if "other" in fields:
                other = fields.get("other")
                field_overwrite_values = self.overwrite_values.get(other, [])
                self._overwrite_value(fields, new_values, field_overwrite_values, key)

        return fields

    @staticmethod
    def _overwrite_values_in_list(fields, new_values, overwrites, key):
        for index, value in enumerate(new_values):
            if value in overwrites:
                fields[key][index] = overwrites.get(value)

    @staticmethod
    def _overwrite_value(fields, new_values, overwrites, key):
        if new_values in overwrites:
            fields[key] = overwrites.get(new_values)

    @staticmethod
    def _format_values(values: List[str]) -> str:
        return ", ".join(values)

    def _overwrite_messages(self, error, *fields):
        for field in fields:
            if field in self.overwrite_messages:
                error = self.overwrite_messages[field]

        return error

    def _overwrite_fields(self, fields):
        for overwrite_field, overwrite_value in self.overwrite_fields.items():
            for key, value in fields.items():
                value = self._convert_field_to_wildcard_field(value)
                if overwrite_field == value:
                    fields[key] = overwrite_value

        return fields

    def _convert_field_to_wildcard_field(self, field) -> str:
        split_fields = str(field).split(self.config.FIELD_DELIMITER)
        for index, field in enumerate(split_fields):
            if field.isnumeric():
                split_fields[index] = self.config.FIELD_WILD_CARD

        return self.config.FIELD_DELIMITER.join(split_fields)

    def _flatten_output(self, output: dict):
        for field, errors in output.items():
            for error in errors:
                self._flat_list.append(error)

    def _get_field_value(self, field) -> Any:
        return get_field_value(self.data, field)

    def field_rules(self, field: str) -> List[str]:
        return self._split_rules(self.rules.get(field))

    @staticmethod
    def valid_email(value: Any) -> bool:
        return rls.EmailRule.valid_email(value)

    @staticmethod
    def valid_url(value: Any) -> bool:
        return rls.UrlRule.valid_url(value)

    @staticmethod
    def valid_ip(value: Any) -> bool:
        return rls.IpRule.valid_ip(value)

    @staticmethod
    def valid_uuid4(value: Any) -> bool:
        return rls.Uuid4Rule.valid_uuid4(value)

    @staticmethod
    def valid_string(value: Any) -> bool:
        return rls.StringRule.valid_string(value)

    @staticmethod
    def valid_integer(value: Any) -> bool:
        return rls.IntegerRule.valid_integer(value)

    @staticmethod
    def valid_float(value: Any) -> bool:
        return rls.FloatRule.valid_float(value)

    @staticmethod
    def valid_decimal(value: Any) -> bool:
        return rls.DecimalRule.valid_decimal(value)

    @staticmethod
    def valid_boolean(value: Any) -> bool:
        return rls.BooleanRule.valid_boolean(value)

    @staticmethod
    def valid_json(value: Any) -> bool:
        return rls.JsonRule.valid_json(value)

    @staticmethod
    def valid_alpha_num(value: Any) -> bool:
        return rls.AlphaNumRule.valid_alpha_num(value)

    @staticmethod
    def valid_alpha_num_space(value: Any) -> bool:
        return rls.AlphaNumSpaceRule.valid_alpha_num_space(value)

    @staticmethod
    def valid_list(value: Any) -> bool:
        return rls.ListRule.valid_list(value)

    @staticmethod
    def valid_dict(value: Any) -> bool:
        return rls.DictRule.valid_dict(value)

    @staticmethod
    def valid_date_time(value: Any, date_time_format: str = None) -> bool:
        return rls.DateTimeRule.valid_date_time(value, date_time_format)



from src.spotlight.errors import (
    MAX_STRING_ERROR,
    MIN_STRING_ERROR,
    IN_ERROR,
    REQUIRED_ERROR,
)
from .validator_test import ValidatorTest


class ListValidationTest(ValidatorTest):
    def setUp(self):
        self.validator.overwrite_messages = {}
        self.validator.overwrite_fields = {}
        self.validator.overwrite_values = {}
        self.rules = {"list.*.test": "max:2", "list.*.test2": "min:2"}
        self.data = {
            "list": [
                {"test": "12", "test2": "1"},
                {"test": "123", "test2": "1"},
                {"test": "1234", "test2": "1"},
            ]
        }

    def test_list_validation_expect_error(self):
        expected = {
            "list.1.test": [MAX_STRING_ERROR.format(field="list.1.test", max=2)],
            "list.2.test": [MAX_STRING_ERROR.format(field="list.2.test", max=2)],
            "list.0.test2": [MIN_STRING_ERROR.format(field="list.0.test2", min=2)],
            "list.1.test2": [MIN_STRING_ERROR.format(field="list.1.test2", min=2)],
            "list.2.test2": [MIN_STRING_ERROR.format(field="list.2.test2", min=2)],
        }

        errors = self.validator.validate(self.data, self.rules)

        self.assertEqual(errors, expected)

    def test_list_validation_flat_expect_error(self):
        expected = [
            MAX_STRING_ERROR.format(field="list.1.test", max=2),
            MAX_STRING_ERROR.format(field="list.2.test", max=2),
            MIN_STRING_ERROR.format(field="list.0.test2", min=2),
            MIN_STRING_ERROR.format(field="list.1.test2", min=2),
            MIN_STRING_ERROR.format(field="list.2.test2", min=2),
        ]

        errors = self.validator.validate(self.data, self.rules, flat=True)

        self.assertEqual(errors, expected)

    def test_list_validation_expect_with_custom_message_error(self):
        new_message = "Hey! The {field} can't be greater than {max} chars!"
        expected = {
            "list.1.test": [new_message.format(field="list.1.test", max=2)],
            "list.2.test": [new_message.format(field="list.2.test", max=2)],
            "list.0.test2": [MIN_STRING_ERROR.format(field="list.0.test2", min=2)],
            "list.1.test2": [MIN_STRING_ERROR.format(field="list.1.test2", min=2)],
            "list.2.test2": [MIN_STRING_ERROR.format(field="list.2.test2", min=2)],
        }
        self.validator.overwrite_messages = {"list.*.test.max": new_message}

        errors = self.validator.validate(self.data, self.rules)

        self.assertEqual(errors, expected)

    def test_list_validation_expect_with_custom_fields_error(self):
        expected = {
            "list.1.test": [MAX_STRING_ERROR.format(field="list.1.test", max=2)],
            "list.2.test": [MAX_STRING_ERROR.format(field="list.2.test", max=2)],
            "list.0.test2": [MIN_STRING_ERROR.format(field="custom", min=2)],
            "list.1.test2": [MIN_STRING_ERROR.format(field="custom", min=2)],
            "list.2.test2": [MIN_STRING_ERROR.format(field="custom", min=2)],
        }
        self.validator.overwrite_fields = {"list.*.test2": "custom"}

        errors = self.validator.validate(self.data, self.rules)

        self.assertEqual(errors, expected)

    def test_list_validation_with_values_message_expect_error_with_new_values(self):
        rules = {"list.*.test": "max:2", "list.*.test2": "in:val1,val2,val3"}
        new_values = "new1, new2, new3"
        expected = {
            "list.1.test": [MAX_STRING_ERROR.format(field="list.1.test", max=2)],
            "list.2.test": [MAX_STRING_ERROR.format(field="list.2.test", max=2)],
            "list.0.test2": [IN_ERROR.format(field="list.0.test2", values=new_values)],
            "list.1.test2": [IN_ERROR.format(field="list.1.test2", values=new_values)],
            "list.2.test2": [IN_ERROR.format(field="list.2.test2", values=new_values)],
        }
        self.validator.overwrite_values = {
            "list.*.test2": {"val1": "new1", "val2": "new2", "val3": "new3"}
        }

        errors = self.validator.validate(self.data, rules)

        self.assertEqual(errors, expected)

    def test_list_validation_required_expect_error(self):
        rules = {"list.*.test": "required|max:2", "list.*.test2": "min:2"}
        data = {
            "list": [
                {"test": "12", "test2": "1"},
                {"test2": "1"},
                {"test": "1234", "test2": "1"},
            ]
        }
        expected = {
            "list.1.test": [REQUIRED_ERROR.format(field="list.1.test")],
            "list.2.test": [MAX_STRING_ERROR.format(field="list.2.test", max=2)],
            "list.0.test2": [MIN_STRING_ERROR.format(field="list.0.test2", min=2)],
            "list.1.test2": [MIN_STRING_ERROR.format(field="list.1.test2", min=2)],
            "list.2.test2": [MIN_STRING_ERROR.format(field="list.2.test2", min=2)],
        }

        errors = self.validator.validate(data, rules)

        self.assertEqual(errors, expected)

    def test_list_validation_with_optional_field_expect_no_error(self):
        rules = {"list.*.test": "max:2", "list.*.test2": "min:2"}
        data = {"list": [{}, {}, {}]}
        expected = {}

        errors = self.validator.validate(data, rules)

        self.assertEqual(errors, expected)

    def test_list_validation_with_required_field_expect_error(self):
        rules = {"list.*.test": "required|max:2", "list.*.test2": "min:2"}
        data = {"list": [{}, {}, {}]}
        expected = {
            "list.0.test": [REQUIRED_ERROR.format(field="list.0.test")],
            "list.1.test": [REQUIRED_ERROR.format(field="list.1.test")],
            "list.2.test": [REQUIRED_ERROR.format(field="list.2.test")],
        }

        errors = self.validator.validate(data, rules)

        self.assertEqual(errors, expected)

    def test_list_validation_with_nested_dict_expect_no_error(self):
        rules = {
            "players": "required|list|min:2",
            "hello": "required|dict",
            "hello.test": "required|string",
            "players.*.name": "required|min:3|max:25",
            "players.*.unit": "required|dict",
            "players.*.unit.name": "required|string|max:3",
        }
        data = {
            "hello": {"test": "some_value"},
            "players": [
                {"name": "Test 1", "unit": {"name": "hi"}},
                {"name": "Test 2", "unit": {"name": "hi"}},
            ],
        }
        expected = {}

        errors = self.validator.validate(data, rules)

        self.assertEqual(errors, expected)

    def test_list_validation_with_empty_list_expect_no_error(self):
        rules = {"list.*.test": "max:2", "list.*.test2": "min:2"}
        data = {"list": []}
        expected = {}

        errors = self.validator.validate(data, rules)

        self.assertEqual(errors, expected)

    def test_list_validation_with_dict_expect_no_error(self):
        rules = {"list.*.test": "max:2", "list.*.test2": "min:2"}
        data = {"list": {}}
        expected = {}

        errors = self.validator.validate(data, rules)

        self.assertEqual(errors, expected)

    def test_list_validation_without_key_expect_no_error(self):
        rules = {
            "start_at": "string|min:10",
            "end_at": "string|min:10",
            "ticket_universe.size": "integer",
            "ticket_universe.positions.*": "string",
        }
        data = {
            "start_at": "2017:03:22",
            "end_at": "2028:03:26",
            "ticket_universe": {
                "size": 99,
                "positions": ["fixed:A", "ranged:1:2", "binary", "alpha", "numeric"],
            },
        }
        expected = {}

        errors = self.validator.validate(data, rules)

        self.assertEqual(errors, expected)



from decimal import Decimal

from src.spotlight.errors import SIZE_ERROR
from .validator_test import ValidatorTest


class SizeTest(ValidatorTest):
    def setUp(self):
        self.field = "test"

    def test_size_rule_with_invalid_string_length_expect_error(self):
        rules = {"test": "size:5"}
        data = {"test": "1234"}
        expected = SIZE_ERROR.format(field=self.field, size=5)

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_size_rule_with_valid_string_length_expect_no_error(self):
        rules = {"test": "size:5"}
        data = {"test": "12345"}
        expected = None

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)

    def test_size_rule_with_empty_string_expect_error(self):
        rules = {"test": "size:5"}
        data = {"test": ""}
        expected = SIZE_ERROR.format(field=self.field, size=5)

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_size_rule_with_invalid_integer_expect_error(self):
        rules = {"test": "size:5"}
        data = {"test": 4}
        expected = SIZE_ERROR.format(field=self.field, size=5)

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_size_rule_with_valid_integer_expect_no_error(self):
        rules = {"test": "size:5"}
        data = {"test": 5}
        expected = {}

        errors = self.validator.validate(data, rules)

        self.assertEqual(errors, expected)

    def test_size_rule_with_invalid_float_expect_error(self):
        rules = {"test": "size:0.5"}
        data = {"test": 0.4}
        expected = SIZE_ERROR.format(field=self.field, size=0.5)

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_size_rule_with_valid_float_expect_no_error(self):
        rules = {"test": "size:0.5"}
        data = {"test": 0.5}
        expected = {}

        errors = self.validator.validate(data, rules)

        self.assertEqual(errors, expected)

    def test_size_rule_with_list_with_invalid_size_expect_error(self):
        rules = {"test": "size:5"}
        data = {"test": [1, 2, 3, 4]}
        expected = SIZE_ERROR.format(field=self.field, size=5)

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_size_rule_with_list_with_valid_size_expect_no_error(self):
        rules = {"test": "size:5"}
        data = {"test": [1, 2, 3, 4, 5]}
        expected = {}

        errors = self.validator.validate(data, rules)

        self.assertEqual(errors, expected)

    def test_size_rule_with_dict_with_invalid_size_expect_error(self):
        rules = {"test": "size:5"}
        data = {"test": {"test1": 1, "test2": 2, "test3": 3, "test4": 4}}
        expected = SIZE_ERROR.format(field=self.field, size=5)

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_size_rule_with_dict_with_valid_size_expect_no_error(self):
        rules = {"test": "size:5"}
        data = {"test": {"test1": 1, "test2": 2, "test3": 3, "test4": 4, "test5": 5}}
        expected = {}

        errors = self.validator.validate(data, rules)

        self.assertEqual(errors, expected)

    def test_size_rule_with_int_value_and_float_rule_expect_error(self):
        rules = {"test": "size:1.1"}
        data = {"test": 1}
        expected = SIZE_ERROR.format(field=self.field, size=1.1)

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_size_rule_with_decimal_value_with_valid_size_expect_no_error(self):
        rules = {"test": "size:5.2"}
        data = {"test": Decimal("5.2")}
        expected = {}

        errors = self.validator.validate(data, rules)

        self.assertEqual(errors, expected)

    def test_size_rule_with_decimal_value_with_invalid_size_expect_error(self):
        rules = {"test": "size:1.1"}
        data = {"test": Decimal("1.2")}
        expected = SIZE_ERROR.format(field=self.field, size=1.1)

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)



from src.spotlight.errors import EMAIL_ERROR
from .validator_test import ValidatorTest


class CustomData:
    def __init__(self, email: str):
        self.email = email


class NestedData:
    def __init__(self, data: CustomData):
        self.data = data


class ObjectInputTest(ValidatorTest):
    def test_object_input_with_one_rule_expect_error(self):
        field = "email"
        rules = {"email": "required|email"}
        data = CustomData(email="john.doe@")
        expected = EMAIL_ERROR.format(field=field)

        errors = self.validator.validate(data, rules)
        errs = errors.get(field)

        self.assertEqual(errs[0], expected)

    def test_nested_object_input_with_one_rule_expect_error(self):
        field = "data.email"
        rules = {"data.email": "required|email"}
        data = CustomData(email="john.doe@")
        parent_data = NestedData(data)

        expected = {field: [EMAIL_ERROR.format(field=field)]}

        errors = self.validator.validate(parent_data, rules)

        self.assertEqual(errors, expected)



from decimal import Decimal

from src.spotlight.errors import MAX_STRING_ERROR, MAX_ERROR, MAX_ITEMS_ERROR
from .validator_test import ValidatorTest


class MaxTest(ValidatorTest):
    def setUp(self):
        self.field = "test"

    def test_max_rule_with_invalid_string_length_expect_error(self):
        rules = {"test": "max:5"}
        data = {"test": "123456"}
        expected = MAX_STRING_ERROR.format(field=self.field, max=5)

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_max_rule_with_valid_string_length_expect_no_error(self):
        rules = {"test": "max:5"}
        data = {"test": "12345"}
        expected = None

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)

    def test_max_rule_with_invalid_integer_expect_error(self):
        rules = {"test": "max:5"}
        data = {"test": 6}
        expected = MAX_ERROR.format(field=self.field, max=5)

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_max_rule_with_invalid_decimal_expect_error(self):
        rules = {"test": "max:5"}
        data = {"test": Decimal("6")}
        expected = MAX_ERROR.format(field=self.field, max=5)

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_max_rule_with_invalid_float_expect_error(self):
        rules = {"test": "max:5.5"}
        data = {"test": 5.6}
        expected = MAX_ERROR.format(field=self.field, max=5.5)

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_max_rule_with_list_size_expect_error(self):
        rules = {"test": "max:5"}
        data = {"test": [1, 2, 3, 4, 5, 6]}
        expected = MAX_ITEMS_ERROR.format(field=self.field, max=5)

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_max_rule_with_none_expect_no_error(self):
        rules = {"test": "string|max:255"}
        data = {"test": None}
        expected = None

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)



from src.spotlight.errors import FLOAT_ERROR
from .validator_test import ValidatorTest


class FloatTest(ValidatorTest):
    def setUp(self):
        self.field = "test"
        self.float_error = FLOAT_ERROR.format(field=self.field)

    def test_float_rule_with_string_expect_error(self):
        rules = {"test": "float"}
        data = {"test": "test"}
        expected = self.float_error

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_float_rule_with_ten_expect_no_error(self):
        rules = {"test": "float"}
        data = {"test": 10.0}
        expected = None

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)

    def test_valid_float_with_invalid_values_expect_false(self):
        invalid_floats = ["0", "1", -1, 0, 1, 2, [], {}, True, False, "test"]

        for invalid_float in invalid_floats:
            actual = self.validator.valid_float(invalid_float)
            self.assertEqual(actual, False)

    def test_valid_float_with_valid_values_expect_true(self):
        valid_floats = [-1.0, 0.0, 1.0, 5.0, 10.0]

        for valid_float in valid_floats:
            actual = self.validator.valid_float(valid_float)
            self.assertEqual(actual, True)



from src.spotlight.errors import DICT_ERROR, REQUIRED_ERROR
from .validator_test import ValidatorTest


class DictTest(ValidatorTest):
    def setUp(self):
        self.field = "test"
        self.rules = {"test": "dict"}
        self.dict_error = DICT_ERROR.format(field=self.field)

    def test_dict_rule_with_set_expect_error(self):
        data = {"test": set()}
        expected = self.dict_error

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_dict_rule_with_dict_expect_no_error(self):
        data = {"test": {}}
        expected = None

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)

    def test_dict_rule_and_required_with_empty_dict_expect_error(self):
        rules = {"test": "required|dict"}
        data = {"test": {}}
        expected = REQUIRED_ERROR.format(field=self.field)

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_dict_rule_and_required_with_non_empty_dict_expect_no_error(self):
        rules = {"test": "required|dict"}
        data = {"test": {"field": "value"}}
        expected = None

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)

    def test_valid_dict_with_invalid_values_expect_false(self):
        invalid_dicts = ["0", "1", -1, 0, 1, 2, [], True, False, list(), set()]

        for invalid_dict in invalid_dicts:
            actual = self.validator.valid_dict(invalid_dict)
            self.assertEqual(actual, False)

    def test_valid_dict_with_valid_values_expect_true(self):
        valid_dicts = [{}, dict()]

        for valid_dict in valid_dicts:
            actual = self.validator.valid_dict(valid_dict)
            self.assertEqual(actual, True)



import ipaddress
import json
import re
from datetime import datetime
from decimal import Decimal
from json import JSONDecodeError
from typing import Any, Tuple, List
from uuid import UUID
from abc import ABC, abstractmethod

from . import errors, config
from .exceptions import (
    RuleNameAlreadyExistsError,
    AttributeNotImplementedError,
    InvalidDateTimeFormat,
)
from .utils import missing, equal, empty, regex_match


class Rule(ABC):
    name = NotImplemented
    implicit = False
    stop = False

    subclasses = []

    def __init__(self):
        self.message_fields = {}

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        if cls.name is NotImplemented:
            raise AttributeNotImplementedError("name", cls.__name__)
        if cls.name in [subclass.name for subclass in cls.subclasses]:
            raise RuleNameAlreadyExistsError(cls.name)

        cls.subclasses.append(cls)

    @abstractmethod
    def passes(self, field: str, value: Any, parameters: List[str], validator) -> bool:
        """
        Tests if the field that is being validated passes the rule.

        Parameters
        ----------
        field : str
            The name of the field that is being validated.
        value : Any
            The value of the field that is being validated.
        parameters : list
            A list of rule parameters.
        validator : Validator
            Instance of the validator. Can be used to access data and rules.
            Useful for more advanced rules that are dependent on other fields.

        Returns
        -------
        bool
            Returns a boolean that indicates if the field that is being
            validated passed the rule.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def message(self) -> str:
        raise NotImplementedError


class RequiredRule(Rule):
    """Required field"""

    name = "required"
    implicit = True
    stop = True

    def passes(self, field: str, value: Any, parameters: List[str], validator) -> bool:
        self.message_fields = dict(field=field)

        return not missing(validator.data, field) and not empty(value)

    @property
    def message(self) -> str:
        return errors.REQUIRED_ERROR


class RequiredWithoutRule(Rule):
    """Required if other field is not present"""

    name = "required_without"
    implicit = True
    stop = True

    def passes(self, field: str, value: Any, parameters: List[str], validator) -> bool:
        other_fields = parameters
        data = validator.data
        self.message_fields = dict(field=field, other=", ".join(other_fields))

        if missing(data, field) and any(missing(data, o) for o in other_fields):
            return False

        return True

    @property
    def message(self) -> str:
        return errors.REQUIRED_WITHOUT_ERROR


class RequiredWithRule(Rule):
    """Required with other field"""

    name = "required_with"
    implicit = True
    stop = True

    def passes(self, field: str, value: Any, parameters: List[str], validator) -> bool:
        other_fields = parameters
        data = validator.data
        self.message_fields = dict(field=field, other=", ".join(other_fields))

        if missing(data, field) and any(not missing(data, o) for o in other_fields):
            return False

        return True

    @property
    def message(self) -> str:
        return errors.REQUIRED_WITH_ERROR


class RequiredIfRule(Rule):
    """Required if other field equals certain value"""

    name = "required_if"
    implicit = True
    stop = True

    def passes(self, field: str, value: Any, parameters: List[str], validator) -> bool:
        other, val = parameters
        data = validator.data
        other_val = data.get(other)
        self.message_fields = dict(field=field, other=other, value=val)

        if missing(data, field) and equal(val, other_val):
            return False

        return True

    @property
    def message(self) -> str:
        return errors.REQUIRED_IF_ERROR


class RequiredUnlessRule(Rule):
    """Required unless other field equals certain value"""

    name = "required_unless"
    implicit = True
    stop = True

    def passes(self, field: str, value: Any, parameters: List[str], validator) -> bool:
        other, val = parameters
        data = validator.data
        other_val = data.get(other)
        self.message_fields = dict(field=field, other=other, value=val)

        if missing(data, field) and not equal(val, other_val):
            return False

        return True

    @property
    def message(self) -> str:
        return errors.REQUIRED_UNLESS_ERROR


class NotWithRule(Rule):
    """Not with other field"""

    name = "not_with"
    stop = True

    def passes(self, field: str, value: Any, parameters: List[str], validator) -> bool:
        other = parameters[0]
        data = validator.data
        self.message_fields = dict(field=field, other=other)

        if not missing(data, field) and not missing(data, other):
            return False

        return True

    @property
    def message(self) -> str:
        return errors.NOT_WITH_ERROR


class FilledRule(Rule):
    """Not empty when present"""

    name = "filled"
    implicit = True
    stop = True

    def passes(self, field: str, value: Any, parameters: List[str], validator) -> bool:
        self.message_fields = dict(field=field)

        if not missing(validator.data, field) and empty(value):
            return False

        return True

    @property
    def message(self) -> str:
        return errors.FILLED_ERROR


class EmailRule(Rule):
    """Valid email"""

    name = "email"
    _regex = re.compile(
        r"^[a-zA-Z0-9.!#$%&’*+/=?^_`{|}~-]+@[a-zA-Z0-9-]+(?:\.[a-zA-Z0-9-]+)*$"
    )

    def passes(self, field: str, value: Any, parameters: List[str], validator) -> bool:
        self.message_fields = dict(field=field)

        return self.valid_email(value)

    @property
    def message(self) -> str:
        return errors.EMAIL_ERROR

    @staticmethod
    def valid_email(email) -> bool:
        return regex_match(EmailRule._regex, email)


class UrlRule(Rule):
    """Valid URL"""

    name = "url"
    _regex = re.compile(
        r"^(?:http|ftp)s?://"  # http:// or https://
        r"(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|"  # domain...
        r"localhost|"  # localhost...
        r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"  # ...or ip
        r"(?::\d+)?"  # optional port
        r"(?:/?|[/?]\S+)$",
        re.IGNORECASE,
    )

    def passes(self, field: str, value: Any, parameters: List[str], validator) -> bool:
        self.message_fields = dict(field=field)

        return self.valid_url(value)

    @property
    def message(self) -> str:
        return errors.URL_ERROR

    @staticmethod
    def valid_url(url) -> bool:
        return regex_match(UrlRule._regex, url)


class IpRule(Rule):
    """Valid IP"""

    name = "ip"

    def passes(self, field: str, value: Any, parameters: List[str], validator) -> bool:
        self.message_fields = dict(field=field)

        return self.valid_ip(value)

    @property
    def message(self) -> str:
        return errors.IP_ERROR

    @staticmethod
    def valid_ip(ip) -> bool:
        if not StringRule.valid_string(ip) and not IntegerRule.valid_integer(ip):
            return False

        try:
            ipaddress.ip_address(ip)
            return True
        except ValueError:
            return False


class MinRule(Rule):
    """Min length"""

    name = "min"

    def __init__(self):
        super().__init__()
        self.error = None

    def passes(self, field: str, value: Any, parameters: List[str], validator) -> bool:
        min_ = parameters[0]
        self.message_fields = dict(field=field, min=min_)
        self.error = errors.MIN_ERROR
        expected = float(min_)

        if isinstance(value, str):
            self.error = errors.MIN_STRING_ERROR
            return len(value) >= expected
        elif isinstance(value, list) or isinstance(value, dict):
            self.error = errors.MIN_ITEMS_ERROR
            return len(value) >= expected
        elif isinstance(value, int):
            return value >= expected
        elif isinstance(value, float):
            return value >= expected
        elif isinstance(value, Decimal):
            expected = Decimal(min_)
            return value >= expected

        return False

    @property
    def message(self) -> str:
        return self.error


class MaxRule(Rule):
    """Max length"""

    name = "max"

    def __init__(self):
        super().__init__()
        self.error = None

    def passes(self, field: str, value: Any, parameters: List[str], validator) -> bool:
        max_ = parameters[0]
        self.message_fields = dict(field=field, max=max_)
        self.error = errors.MAX_ERROR
        expected = float(max_)

        if isinstance(value, str):
            self.error = errors.MAX_STRING_ERROR
            return len(value) <= expected
        elif isinstance(value, list) or isinstance(value, dict):
            self.error = errors.MAX_ITEMS_ERROR
            return len(value) <= expected
        elif isinstance(value, int):
            return value <= expected
        elif isinstance(value, float):
            return value <= expected
        elif isinstance(value, Decimal):
            expected = Decimal(max_)
            return value <= expected

        return False

    @property
    def message(self) -> str:
        return self.error


class InRule(Rule):
    """
        In: The field under validation must be included in the given list
        of values
    """

    name = "in"

    def passes(self, field: str, value: Any, parameters: List[str], validator) -> bool:
        self.message_fields = dict(field=field, values=parameters)

        return str(value) in parameters

    @property
    def message(self) -> str:
        return errors.IN_ERROR


class AlphaNumRule(Rule):
    """Only letters and numbers"""

    name = "alpha_num"
    _regex = re.compile(r"^[a-zA-Z0-9]+$")

    def passes(self, field: str, value: Any, parameters: List[str], validator) -> bool:
        self.message_fields = dict(field=field)

        return self.valid_alpha_num(value)

    @property
    def message(self) -> str:
        return errors.ALPHA_NUM_ERROR

    @staticmethod
    def valid_alpha_num(value) -> bool:
        return regex_match(AlphaNumRule._regex, value)


class AlphaNumSpaceRule(Rule):
    """Only letters, numbers and spaces"""

    name = "alpha_num_space"
    _regex = re.compile(r"^[a-zA-Z0-9 ]+$")

    def passes(self, field: str, value: Any, parameters: List[str], validator) -> bool:
        self.message_fields = dict(field=field)

        return self.valid_alpha_num_space(value)

    @property
    def message(self) -> str:
        return errors.ALPHA_NUM_SPACE_ERROR

    @staticmethod
    def valid_alpha_num_space(value) -> bool:
        return regex_match(AlphaNumSpaceRule._regex, value)


class StringRule(Rule):
    """Valid string"""

    name = "string"

    def passes(self, field: str, value: Any, parameters: List[str], validator) -> bool:
        self.message_fields = dict(field=field)

        return self.valid_string(value)

    @property
    def message(self) -> str:
        return errors.STRING_ERROR

    @staticmethod
    def valid_string(string) -> bool:
        return type(string) is str


class IntegerRule(Rule):
    """Valid integer"""

    name = "integer"

    def passes(self, field: str, value: Any, parameters: List[str], validator) -> bool:
        self.message_fields = dict(field=field)

        return self.valid_integer(value)

    @property
    def message(self) -> str:
        return errors.INTEGER_ERROR

    @staticmethod
    def valid_integer(integer) -> bool:
        return type(integer) is int


class FloatRule(Rule):
    """Valid float"""

    name = "float"

    def passes(self, field: str, value: Any, parameters: List[str], validator) -> bool:
        self.message_fields = dict(field=field)

        return self.valid_float(value)

    @property
    def message(self) -> str:
        return errors.FLOAT_ERROR

    @staticmethod
    def valid_float(float_) -> bool:
        return type(float_) is float


class DecimalRule(Rule):
    """Valid decimal"""

    name = "decimal"

    def passes(self, field: str, value: Any, parameters: List[str], validator) -> bool:
        self.message_fields = dict(field=field)

        return self.valid_decimal(value)

    @property
    def message(self) -> str:
        return errors.DECIMAL_ERROR

    @staticmethod
    def valid_decimal(value) -> bool:
        return isinstance(value, Decimal)


class BooleanRule(Rule):
    """Valid boolean"""

    name = "boolean"

    def passes(self, field: str, value: Any, parameters: List[str], validator) -> bool:
        self.message_fields = dict(field=field)

        return self.valid_boolean(value)

    @property
    def message(self) -> str:
        return errors.BOOLEAN_ERROR

    @staticmethod
    def valid_boolean(boolean) -> bool:
        return type(boolean) is bool


class ListRule(Rule):
    """Valid list"""

    name = "list"
    stop = True

    def passes(self, field: str, value: Any, parameters: List[str], validator) -> bool:
        self.message_fields = dict(field=field)

        return self.valid_list(value)

    @property
    def message(self) -> str:
        return errors.LIST_ERROR

    @staticmethod
    def valid_list(value) -> bool:
        return type(value) is list


class Uuid4Rule(Rule):
    """Valid uuid4"""

    name = "uuid4"

    def passes(self, field: str, value: Any, parameters: List[str], validator) -> bool:
        self.message_fields = dict(field=field)

        return self.valid_uuid4(value)

    @property
    def message(self) -> str:
        return errors.UUID4_ERROR

    @staticmethod
    def valid_uuid4(uuid) -> bool:
        if type(uuid) is UUID:
            uuid = str(uuid)
        try:
            val = UUID(uuid, version=4)
        except (TypeError, ValueError, AttributeError):
            return False

        return str(val) == uuid


class JsonRule(Rule):
    """Valid json"""

    name = "json"

    def passes(self, field: str, value: Any, parameters: List[str], validator) -> bool:
        self.message_fields = dict(field=field)

        return self.valid_json(value)

    @property
    def message(self) -> str:
        return errors.JSON_ERROR

    @staticmethod
    def valid_json(value) -> bool:
        try:
            json.loads(value)
            return True
        except (TypeError, JSONDecodeError):
            return False


class AcceptedRule(Rule):
    """The field must be yes, on, 1, or true"""

    name = "accepted"

    def passes(self, field: str, value: Any, parameters: List[str], validator) -> bool:
        accepted_values = ["yes", "on", 1, True]
        self.message_fields = dict(field=field)

        return value in accepted_values

    @property
    def message(self) -> str:
        return errors.ACCEPTED_ERROR


class StartsWithRule(Rule):
    """The field under validation must start with one of the given values."""

    name = "starts_with"

    def passes(self, field: str, value: Any, parameters: List[str], validator) -> bool:
        self.message_fields = dict(field=field, values=parameters)

        return any([str(value).startswith(rule_val) for rule_val in parameters])

    @property
    def message(self) -> str:
        return errors.STARTS_WITH_ERROR


class DictRule(Rule):
    """Valid dict"""

    name = "dict"

    def passes(self, field: str, value: Any, parameters: List[str], validator) -> bool:
        self.message_fields = dict(field=field)

        return self.valid_dict(value)

    @property
    def message(self) -> str:
        return errors.DICT_ERROR

    @staticmethod
    def valid_dict(value) -> bool:
        return type(value) is dict


class DateTimeRule(Rule):
    """
    Valid date/time matching the default 'YYYY-MM-DD hh:mm:ss' format, or a
    custom specified format.
    """

    name = "date_time"
    stop = True
    default_format = config.DEFAULT_DATE_TIME_FORMAT

    _regex = re.compile(r"^\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}$")

    def passes(self, field: str, value: Any, parameters: List[str], validator) -> bool:
        supplied_format = parameters[0] if parameters else None
        _date_time_format = supplied_format or DateTimeRule.default_format
        self.message_fields = dict(field=field, format=_date_time_format)

        return self.valid_date_time(value, supplied_format)

    @property
    def message(self) -> str:
        return errors.DATE_TIME_ERROR

    @staticmethod
    def valid_date_time(value: Any, date_time_format: str = None) -> bool:
        if isinstance(value, datetime):
            return True

        if not date_time_format and not regex_match(DateTimeRule._regex, value):
            return False

        try:
            datetime.strptime(value, date_time_format or DateTimeRule.default_format)
        except (ValueError, TypeError):
            return False

        return True


class BeforeRule(Rule):
    """Date/time that must occur before another date/time."""

    name = "before"

    def passes(self, field: str, value: Any, parameters: List[str], validator) -> bool:
        supplied_field_or_format = parameters[0] if parameters else None
        before_date, before_format = self.date_and_format(field, field, validator)
        after_date, after_format = self.date_and_format(
            field, supplied_field_or_format, validator
        )
        self.message_fields = dict(field=field, other=supplied_field_or_format)

        return before_date < after_date

    @staticmethod
    def date_and_format(
        field: str, field_or_date_time: Any, validator
    ) -> Tuple[datetime, str]:
        after_format = DateTimeRule.default_format

        # First try the value as a datetime string with the default format. If
        # it fails, try and find out if a datetime format has been specified in
        # the rule set.
        try:
            after_date = datetime.strptime(field_or_date_time, after_format)
        except (ValueError, TypeError):
            after_format = BeforeRule.date_time_field_format(
                field_or_date_time, validator
            )
            value = validator.data.get(field_or_date_time)

            if isinstance(value, datetime):
                after_date = value
            else:
                try:
                    after_date = datetime.strptime(value, after_format)
                except (ValueError, TypeError):
                    after_format = BeforeRule.date_time_field_format(field, validator)
                    try:
                        after_date = datetime.strptime(field_or_date_time, after_format)
                    except (ValueError, TypeError):
                        raise InvalidDateTimeFormat

        return after_date, after_format

    @staticmethod
    def date_time_field_format(field, validator) -> str:
        date_time_format = None
        if field in validator.rules:
            rules = validator.field_rules(field)
            for name, parameters in validator.rule_iterator(rules):
                if name == DateTimeRule.name and parameters:
                    date_time_format = parameters[0]

        return date_time_format or DateTimeRule.default_format

    @property
    def message(self) -> str:
        return errors.BEFORE_ERROR


class AfterRule(Rule):
    """Date/time that must occur after another date/time."""

    name = "after"

    def passes(self, field: str, value: Any, parameters: List[str], validator) -> bool:
        supplied_field_or_format = parameters[0] if parameters else None
        after_date, after_format = BeforeRule.date_and_format(field, field, validator)
        before_date, before_format = BeforeRule.date_and_format(
            field, supplied_field_or_format, validator
        )
        self.message_fields = dict(field=field, other=supplied_field_or_format)

        return after_date > before_date

    @property
    def message(self) -> str:
        return errors.AFTER_ERROR


class SizeRule(Rule):
    """Size"""

    name = "size"

    def passes(self, field: str, value: Any, parameters: List[str], validator) -> bool:
        size = parameters[0]
        self.message_fields = dict(field=field, size=size)
        expected = float(size)

        if isinstance(value, str):
            return len(value) == expected
        elif isinstance(value, list) or isinstance(value, dict):
            return len(value) == expected
        elif isinstance(value, int):
            return value == expected
        elif isinstance(value, float):
            return value == expected
        elif isinstance(value, Decimal):
            expected = Decimal(size)
            return value == expected

        return False

    @property
    def message(self) -> str:
        return errors.SIZE_ERROR


class EndsWithRule(Rule):
    """The field under validation must end with one of the given values."""

    name = "ends_with"

    def passes(self, field: str, value: Any, parameters: List[str], validator) -> bool:
        self.message_fields = dict(field=field, values=parameters)

        return any([str(value).endswith(rule_val) for rule_val in parameters])

    @property
    def message(self) -> str:
        return errors.ENDS_WITH_ERROR



from src.spotlight.errors import LIST_ERROR, REQUIRED_ERROR
from .validator_test import ValidatorTest


class ListTest(ValidatorTest):
    def setUp(self):
        self.field = "test"
        self.rules = {"test": "list"}
        self.list_error = LIST_ERROR.format(field=self.field)

    def test_list_rule_with_integer_expect_error(self):
        data = {"test": 1}
        expected = self.list_error

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_list_rule_with_list_expect_no_error(self):
        data = {"test": []}
        expected = None

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)

    def test_list_rule_and_required_with_empty_list_expect_error(self):
        rules = {"test": "required|list"}
        data = {"test": []}
        expected = REQUIRED_ERROR.format(field=self.field)

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_list_rule_with_boolean_true_expect_error(self):
        data = {"test": True}
        expected = self.list_error

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_valid_list_with_dict_expect_false(self):
        actual = self.validator.valid_list(dict())

        self.assertEqual(actual, False)

    def test_valid_list_with_list_expect_true(self):
        actual = self.validator.valid_list([])

        self.assertEqual(actual, True)



from src.spotlight.errors import REQUIRED_WITH_ERROR
from .validator_test import ValidatorTest


class RequiredWithTest(ValidatorTest):
    def setUp(self):
        self.other_field = "test1"
        self.field = "test2"
        self.required_with_error = REQUIRED_WITH_ERROR.format(
            field=self.field, other=self.other_field
        )
        self.rules = {"test2": "required_with:test1"}

    def test_required_with_rule_with_missing_field_expect_error(self):
        data = {"test1": "hello"}
        expected = self.required_with_error

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_required_with_rule_with_field_present_expect_no_error(self):
        data = {"test1": "hello", "test2": "world"}
        expected = None

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)

    def test_required_with_rule_with_boolean_true_expect_no_error(self):
        data = {"test1": True, "test2": "world"}
        expected = None

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)

    def test_required_with_rule_with_boolean_false_expect_no_error(self):
        data = {"test1": False, "test2": "world"}
        expected = None

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)

    def test_required_with_rule_with_multi_requirement_and_missing_field_expect_error(
        self
    ):
        field = "test5"
        rules = {"test5": "required_with:test1,test2,test3,test4"}
        data = {"test2": "not.missing", "test4": "not.missing"}
        expected = REQUIRED_WITH_ERROR.format(
            field=field, other="test1, test2, test3, test4"
        )

        errors = self.validator.validate(data, rules)
        errs = errors.get(field)

        self.assertEqual(errs[0], expected)

    def test_required_with_rule_with_all_present_expect_no_error(self):
        rules = {"test5": "required_with:test1,test2,test3,test4"}
        data = {
            "test1": "test",
            "test2": "test",
            "test3": "test",
            "test4": "test",
            "test5": "test",
        }
        expected = None

        errors = self.validator.validate(data, rules)
        errs = errors.get("test5")

        self.assertEqual(errs, expected)



from .validator_test import ValidatorTest


class FlatParamTest(ValidatorTest):
    def setUp(self):
        self.rules = {
            "email": "email|max:4",
            "password": "min:8",
            "nested1.field": "min:5",
            "nested2.*": "min:5",
            "nested3.*.field": "min:5",
            "nested4.nested.field": "min:5",
        }
        self.data = {
            "email": "this.is.not.a.valid.email",
            "password": "test",
            "nested1": {"field": "test"},
            "nested2": ["test", "valid", "test"],
            "nested3": [{"field": "test"}, {"field": "valid"}, {"field": "test"}],
            "nested4": {"nested": {"field": "test"}},
        }

    def test_flat_param_true_expect_list_of_errors(self):
        errors = self.validator.validate(self.data, self.rules, flat=True)

        self.assertEqual(isinstance(errors, list), True)
        self.assertEqual(len(errors), 9)

    def test_flat_param_false_expect_dict_of_errors(self):
        errors = self.validator.validate(self.data, self.rules, flat=False)

        self.assertEqual(isinstance(errors, dict), True)
        self.assertEqual(len(errors), 8)
        self.assertEqual(len(errors.get("email")), 2)
        self.assertEqual(len(errors.get("password")), 1)
        self.assertEqual(len(errors.get("nested1.field")), 1)
        self.assertEqual(len(errors.get("nested2.0")), 1)
        self.assertEqual(len(errors.get("nested2.2")), 1)
        self.assertEqual(len(errors.get("nested3.0.field")), 1)
        self.assertEqual(len(errors.get("nested3.2.field")), 1)
        self.assertEqual(len(errors.get("nested4.nested.field")), 1)



from src.spotlight.errors import IP_ERROR
from .validator_test import ValidatorTest


class IpTest(ValidatorTest):
    def test_ip_rule_with_invalid_ips_expect_errors(self):
        rules = {"ip1": "ip", "ip2": "ip", "ip3": "ip"}
        data = {
            "ip1": "this.is.not.valid",
            "ip2": "255.255.255.256",
            "ip3": "www.google.com",
        }

        errors = self.validator.validate(data, rules)

        self.assertEqual(len(errors.items()), 3)
        for field, errs in errors.items():
            expected = IP_ERROR.format(field=field)
            self.assertEqual(errs[0], expected)

    def test_ip_rule_with_valid_ips_expect_no_errors(self):
        rules = {"ip1": "ip", "ip2": "ip", "ip3": "ip", "ip4": "ip", "ip5": "ip"}
        data = {
            "ip1": "192.168.1.1",
            "ip2": "0.0.0.0",
            "ip3": "255.255.255.255",
            "ip4": 3232235777,
            "ip5": "2001:0db8:85a3:0000:0000:8a2e:0370:7334",
        }
        expected = None

        errors = self.validator.validate(data, rules)

        self.assertEqual(len(errors.items()), 0)
        for field, errs in errors.items():
            self.assertEqual(errs, expected)

    def test_valid_ip_with_boolean_true_expect_false(self):
        valid_ip = self.validator.valid_ip(True)

        self.assertEqual(valid_ip, False)

    def test_valid_ip_with_integer_expect_true(self):
        # String value:  192.168.1.1
        # Binary:        11000000 . 10101000 . 00000001 . 00000001
        # Integer:       3232235777
        valid_ip = self.validator.valid_ip(3232235777)

        self.assertEqual(valid_ip, True)



from uuid import uuid4, UUID

from src.spotlight.errors import UUID4_ERROR
from .validator_test import ValidatorTest


class Uuid4Test(ValidatorTest):
    def setUp(self):
        self.field = "test"
        self.uuid4_error = UUID4_ERROR.format(field=self.field)
        self.rules = {"test": "uuid4"}

    def test_uuid4_rule_with_invalid_string_expect_error(self):
        data = {"test": "1234"}
        expected = self.uuid4_error

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_uuid4_rule_with_integer_expect_error(self):
        data = {"test": 1234}
        expected = UUID4_ERROR.format(field=self.field)

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_uuid4_rule_with_boolean_true_expect_error(self):
        data = {"test": True}
        expected = UUID4_ERROR.format(field=self.field)

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_uuid4_rule_with_generated_uuid4_expect_no_error(self):
        rules = {"test": "uuid4"}
        data = {"test": uuid4()}
        expected = None

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)

    def test_uuid4_rule_with_string_expect_no_error(self):
        rules = {"test": "uuid4"}
        data = {"test": "a546ce64-634b-43b3-9c32-3f0353edb294"}
        expected = None

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)

    def test_uuid4_rule_with_empty_string_expect_error(self):
        rules = {"test": "uuid4"}
        data = {"test": ""}
        expected = self.uuid4_error

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_valid_uuid4_with_invalid_values_expect_false(self):
        invalid_uuid4s = [True, False, -1, 0, 1, 2, [], {}, "", "123"]

        for invalid_uuid4 in invalid_uuid4s:
            actual = self.validator.valid_uuid4(invalid_uuid4)
            self.assertEqual(actual, False)

    def test_valid_uuid4_with_valid_values_expect_true(self):
        values = [
            "a546ce64-634b-43b3-9c32-3f0353edb294",
            uuid4(),
            UUID("{12345678-1234-5678-1234-567812345678}", version=4),
            UUID("12345678123456781234567812345678", version=4),
            UUID("urn:uuid:12345678-1234-5678-1234-567812345678", version=4),
            UUID(bytes=b"\x12\x34\x56\x78"*4, version=4),
            UUID(
                bytes_le=b"\x78\x56\x34\x12\x34\x12\x78\x56\x12\x34\x56\x78\x12\x34\x56\x78",
                version=4,
            ),
            UUID(
                fields=(0x12345678, 0x1234, 0x5678, 0x12, 0x34, 0x567812345678),
                version=4,
            ),
            UUID(int=0x12345678123456781234567812345678, version=4),
        ]

        for value in values:
            actual = self.validator.valid_uuid4(value)
            self.assertEqual(actual, True)



from src.spotlight.errors import ENDS_WITH_ERROR
from .validator_test import ValidatorTest


class EndsWithTest(ValidatorTest):
    def setUp(self):
        self.field = "test"
        self.in_error = ENDS_WITH_ERROR.format(
            field=self.field, values="val0, val1, val2"
        )

    def test_ends_with_rule_with_invalid_value_expect_error(self):
        data = {"test": "valueval3"}
        rules = {"test": "ends_with:val0,val1,val2"}
        expected = self.in_error

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_ends_with_rule_with_valid_value_expect_no_error(self):
        data = {"test": "valueval1"}
        rules = {"test": "ends_with:val0,val1,val2"}
        expected = None

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)

    def test_ends_with_rule_with_valid_integer_expect_no_error(self):
        data = {"test": 5555555316}
        rules = {"test": "ends_with:111,316,743"}
        expected = None

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)



from src.spotlight.errors import BOOLEAN_ERROR
from .validator_test import ValidatorTest


class BooleanTest(ValidatorTest):
    def test_boolean_rule_with_invalid_values_expect_error(self):
        rules = {"test1": "boolean", "test2": "boolean"}
        data = {"test1": "test", "test2": 1}

        errors = self.validator.validate(data, rules)

        self.assertEqual(len(errors.items()), 2)
        for field, errs in errors.items():
            expected = BOOLEAN_ERROR.format(field=field)
            self.assertEqual(errs[0], expected)

    def test_boolean_rule_with_valid_values_expect_no_error(self):
        rules = {"test1": "boolean", "test2": "boolean"}
        data = {"test1": True, "test2": False}
        expected = {}

        errors = self.validator.validate(data, rules)

        self.assertEqual(errors, expected)

    def test_valid_boolean_with_invalid_values_expect_false(self):
        invalid_booleans = ["0", "1", -1, 0, 1, 2, [], {}]

        for invalid_boolean in invalid_booleans:
            actual = self.validator.valid_boolean(invalid_boolean)
            self.assertEqual(actual, False)

    def test_valid_boolean_with_valid_values_expect_true(self):
        valid_booleans = [True, False]

        for valid_boolean in valid_booleans:
            actual = self.validator.valid_boolean(valid_boolean)
            self.assertEqual(actual, True)



from src.spotlight.errors import STRING_ERROR
from .validator_test import ValidatorTest


class StringTest(ValidatorTest):
    def setUp(self):
        self.field = "test"
        self.string_error = STRING_ERROR.format(field=self.field)

    def test_string_rule_with_integer_expect_error(self):
        rules = {"test": "string"}
        data = {"test": 1}
        expected = self.string_error

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_string_rule_with_string_expect_no_error(self):
        rules = {"test": "string"}
        data = {"test": "hello world"}
        expected = None

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)

    def test_valid_string_with_invalid_values_expect_false(self):
        invalid_strings = [True, False, -1, 0, 1, 2, [], {}]

        for invalid_string in invalid_strings:
            actual = self.validator.valid_string(invalid_string)
            self.assertEqual(actual, False)

    def test_valid_string_with_valid_values_expect_true(self):
        valid_strings = ["", "hello", "world", "0", "1", "!"]

        for valid_string in valid_strings:
            actual = self.validator.valid_string(valid_string)
            self.assertEqual(actual, True)



from src.spotlight.errors import NOT_WITH_ERROR
from .validator_test import ValidatorTest


class NotWithTest(ValidatorTest):
    def setUp(self):
        self.field1 = "test1"
        self.field2 = "test2"
        self.not_with_error = NOT_WITH_ERROR.format(
            field=self.field2, other=self.field1
        )
        self.rules = {"test2": "not_with:test1"}

    def test_not_with_rule_with_both_fields_expect_error(self):
        data = {"test1": "hello", "test2": "world"}
        expected = self.not_with_error

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field2)

        self.assertEqual(errs[0], expected)

    def test_not_with_rule_with_first_field_expect_no_error(self):
        data = {"test1": "hello"}
        expected = None

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field2)

        self.assertEqual(errs, expected)

    def test_not_with_rule_with_second_field_expect_no_error(self):
        data = {"test2": "world"}
        expected = None

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field2)

        self.assertEqual(errs, expected)



ACCEPTED_ERROR = "The {field} field must be accepted."
AFTER_ERROR = "The {field} field must be a date/time after {other}."
ALPHA_NUM_ERROR = "The {field} field can only contain letters and numbers."
ALPHA_NUM_SPACE_ERROR = "The {field} field can only contain letters, numbers and spaces."
BEFORE_ERROR = "The {field} field must be a date/time before {other}."
BOOLEAN_ERROR = "The {field} field must be a boolean."
DATE_TIME_ERROR = "The {field} field has to be a valid date/time matching the {format} format."
DECIMAL_ERROR = "The {field} field must be a decimal."
DICT_ERROR = "The {field} field must be a dict."
EMAIL_ERROR = "The {field} field has to be a valid email address."
ENDS_WITH_ERROR = "The {field} field must end with one of the following values: {values}."
FILLED_ERROR = "The {field} field must not be empty when it is present."
FLOAT_ERROR = "The {field} field must be a float."
IN_ERROR = "The {field} field must be one of the following values: {values}."
INTEGER_ERROR = "The {field} field must be an integer."
IP_ERROR = "The {field} field has to be a valid IP address."
JSON_ERROR = "The {field} field must be a valid JSON string."
LIST_ERROR = "The {field} field must be a list."
MAX_ERROR = "The {field} field has to have a maximum value of {max}."
MAX_ITEMS_ERROR = "The {field} field cannot contain more than {max} items."
MAX_STRING_ERROR = "The {field} field cannot be longer than {max} characters."
MIN_ERROR = "The {field} field has to have a minimum value of {min}."
MIN_ITEMS_ERROR = "The {field} field has to contain at least {min} items."
MIN_STRING_ERROR = "The {field} field has to be at least {min} characters."
NOT_WITH_ERROR = "The {field} field can't be present when the {other} field is present."
REQUIRED_ERROR = "The {field} field is required."
REQUIRED_IF_ERROR = "The {field} field is required if the {other} field equals {value}."
REQUIRED_UNLESS_ERROR = "The {field} field is required unless the {other} field equals {value}."
REQUIRED_WITH_ERROR = "The {field} field is required if any of the following fields are present: {other}."
REQUIRED_WITHOUT_ERROR = "The {field} field is required if any of the following fields are absent: {other}."
SIZE_ERROR = "The {field} field has to have a size of {size}."
STARTS_WITH_ERROR = "The {field} field must start with one of the following values: {values}."
STRING_ERROR = "The {field} field must be a string."
URL_ERROR = "The {field} field has to be a valid URL."
UUID4_ERROR = "The {field} field must be a valid UUID."



from decimal import Decimal

from src.spotlight.errors import MIN_STRING_ERROR, MIN_ERROR, MIN_ITEMS_ERROR
from .validator_test import ValidatorTest


class MinTest(ValidatorTest):
    def setUp(self):
        self.field = "test"

    def test_min_rule_with_invalid_string_length_expect_error(self):
        rules = {"test": "min:5"}
        data = {"test": "1234"}
        expected = MIN_STRING_ERROR.format(field=self.field, min=5)

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_min_rule_with_valid_string_length_expect_no_error(self):
        rules = {"test": "min:5"}
        data = {"test": "12345"}
        expected = None

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)

    def test_min_rule_with_no_input_expect_no_error(self):
        rules = {"test": "min:5"}
        data = {}
        expected = None

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)

    def test_min_rule_with_empty_string_expect_error(self):
        rules = {"test": "min:5"}
        data = {"test": ""}
        expected = MIN_STRING_ERROR.format(field=self.field, min=5)

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_min_rule_with_invalid_integer_expect_error(self):
        rules = {"test": "min:5"}
        data = {"test": 4}
        expected = MIN_ERROR.format(field=self.field, min=5)

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_min_rule_with_invalid_float_expect_error(self):
        rules = {"test": "min:0.5"}
        data = {"test": 0.4}
        expected = MIN_ERROR.format(field=self.field, min=0.5)

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_min_rule_with_invalid_decimal_expect_error(self):
        rules = {"test": "min:0.5"}
        data = {"test": Decimal("0.4")}
        expected = MIN_ERROR.format(field=self.field, min=0.5)

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_min_rule_with_list_size_expect_error(self):
        rules = {"test": "min:5"}
        data = {"test": [1, 2, 3, 4]}
        expected = MIN_ITEMS_ERROR.format(field=self.field, min=5)

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_min_rule_with_int_value_and_float_rule_expect_error(self):
        rules = {"test": "min:1.1"}
        data = {"test": 1}
        expected = MIN_ERROR.format(field=self.field, min=1.1)

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)



import json

from src.spotlight.errors import JSON_ERROR
from .validator_test import ValidatorTest


class JsonTest(ValidatorTest):
    def setUp(self):
        self.field = "test"
        self.rules = {"test": "json"}
        self.json_error = JSON_ERROR.format(field=self.field)

    def test_json_rule_with_integer_expect_error(self):
        data = {"test": 1}
        expected = self.json_error

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_json_rule_with_boolean_true_expect_error(self):
        data = {"test": True}
        expected = self.json_error

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_json_rule_with_list_expect_error(self):
        data = {"test": []}
        expected = self.json_error

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_json_rule_with_dict_expect_error(self):
        data = {"test": {}}
        expected = self.json_error

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_json_rule_with_json_string_expect_no_error(self):
        data = {"test": '{"key":"value"}'}
        expected = None

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)

    def test_json_rule_with_json_expect_no_error(self):
        json_data = json.dumps({"key": "values"})
        data = {"test": json_data}
        expected = None

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)

    def test_valid_json_with_invalid_values_expect_false(self):
        values = [
            True,
            False,
            -1,
            0,
            1,
            2,
            [],
            {},
            "",
            set(),
            "test",
            "{invalid}",
            '{"invalid"}',
            '{"invalid":}',
            "{invalid: data}",
        ]

        for value in values:
            actual = self.validator.valid_json(value)
            self.assertEqual(actual, False)

    def test_valid_json_with_valid_values_expect_true(self):
        values = [
            json.dumps({"key": "values"}),
            bytearray('{"key": "values"}', "utf-8"),
            str.encode('{"key": "values"}'),
            bytes('{"key": "values"}', "utf-8"),
        ]

        for value in values:
            actual = self.validator.valid_json(value)
            self.assertEqual(actual, True)



from src.spotlight.errors import STARTS_WITH_ERROR
from .validator_test import ValidatorTest


class StartsWithTest(ValidatorTest):
    def setUp(self):
        self.field = "test"
        self.in_error = STARTS_WITH_ERROR.format(
            field=self.field, values="val0, val1, val2"
        )

    def test_starts_with_rule_with_invalid_value_expect_error(self):
        data = {"test": "val3value"}
        rules = {"test": "starts_with:val0,val1,val2"}
        expected = self.in_error

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_starts_with_rule_with_valid_value_expect_no_error(self):
        data = {"test": "val1value"}
        rules = {"test": "starts_with:val0,val1,val2"}
        expected = None

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)

    def test_starts_with_rule_with_valid_integer_expect_no_error(self):
        data = {"test": 3165555555}
        rules = {"test": "starts_with:111,316,555"}
        expected = None

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)



from src.spotlight.errors import INTEGER_ERROR
from .validator_test import ValidatorTest


class IntegerTest(ValidatorTest):
    def setUp(self):
        self.field = "test"
        self.integer_error = INTEGER_ERROR.format(field=self.field)

    def test_integer_rule_with_string_expect_error(self):
        rules = {"test": "integer"}
        data = {"test": "test"}
        expected = self.integer_error

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_integer_rule_with_ten_expect_no_error(self):
        rules = {"test": "integer"}
        data = {"test": 10}
        expected = None

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)

    def test_valid_integers_with_invalid_values_expect_false(self):
        invalid_integers = ["0", "1", -1.0, 1.4, [], {}, True, False, "test"]

        for invalid_integer in invalid_integers:
            actual = self.validator.valid_integer(invalid_integer)
            self.assertEqual(actual, False)

    def test_valid_integers_with_valid_values_expect_true(self):
        valid_integers = [-1, 0, 1, 5, 10]

        for valid_integer in valid_integers:
            actual = self.validator.valid_integer(valid_integer)
            self.assertEqual(actual, True)



from src.spotlight.errors import REQUIRED_WITHOUT_ERROR
from .validator_test import ValidatorTest


class RequiredWithoutTest(ValidatorTest):
    def setUp(self):
        self.other_field = "test1"
        self.field = "test2"
        self.required_without_error = REQUIRED_WITHOUT_ERROR.format(
            field=self.field, other=self.other_field
        )
        self.rules = {"test2": "required_without:test1"}

    def test_required_without_rule_with_missing_field_expect_error(self):
        data = {}
        expected = self.required_without_error

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_required_without_rule_with_field_present_expect_no_error(self):
        data = {"test2": "world"}
        expected = None

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)

    def test_required_without_rule_with_missing_fields_expect_error(self):
        field = "test4"
        rules = {"test4": "required_without:test1,test2,test3"}
        data = {"test2": "not.missing"}
        expected = REQUIRED_WITHOUT_ERROR.format(
            field=field, other="test1, test2, test3"
        )

        errors = self.validator.validate(data, rules)
        errs = errors.get(field)

        self.assertEqual(errs[0], expected)



from src.spotlight.errors import REQUIRED_ERROR
from .validator_test import ValidatorTest


class RequiredTest(ValidatorTest):
    def setUp(self):
        self.field = "test"
        self.required_error = REQUIRED_ERROR.format(field=self.field)
        self.rules = {"test": "required"}

    def test_required_rule_with_no_input_expect_error(self):
        data = {}
        expected = self.required_error

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_required_rule_with_string_expect_no_error(self):
        data = {"test": "this is a string"}
        expected = None

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)

    def test_required_rule_with_integer_one_expect_no_error(self):
        data = {"test": 1}
        expected = None

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)

    def test_required_rule_with_integer_zero_expect_no_error(self):
        data = {"test": 0}
        expected = None

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)

    def test_required_rule_with_boolean_false_expect_no_error(self):
        data = {"test": False}
        expected = None

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)

    def test_required_rule_with_boolean_true_expect_no_error(self):
        data = {"test": True}
        expected = None

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)

    def test_required_rule_with_empty_list_expect_error(self):
        data = {"test": []}
        expected = self.required_error

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_required_rule_with_list_expect_no_error(self):
        data = {"test": ["hello"]}
        expected = None

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)

    def test_required_rule_with_none_expect_error(self):
        data = {"test": None}
        expected = self.required_error

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_required_rule_with_empty_string_expect_error(self):
        data = {"test": ""}
        expected = self.required_error

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_required_rule_with_spaces_expect_error(self):
        data = {"test": "     "}
        expected = self.required_error

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_required_rule_with_empty_dict_expect_error(self):
        data = {"test": dict()}
        expected = self.required_error

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_required_rule_with_dict_expect_no_error(self):
        data = {"test": {"hello": "world"}}
        expected = None

        errors = self.validator.validate(data, self.rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)



from setuptools import setup, find_packages
from collections import OrderedDict

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="spotlight",
    version="1.0.9",
    author="Michiel Doesburg",
    author_email="michiel@moddix.com",
    description="Laravel style data validation for Python.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="spotlight validation validate",
    url="https://github.com/mdoesburg/spotlight",
    project_urls=OrderedDict(
        (
            ("Documentation", "https://github.com/mdoesburg/spotlight"),
            ("Code", "https://github.com/mdoesburg/spotlight"),
        )
    ),
    license="MIT",
    packages=find_packages("src"),
    package_dir={"": "src"},
    include_package_data=True,
    install_requires=[],
)



from src.spotlight.errors import ACCEPTED_ERROR
from .validator_test import ValidatorTest


class AcceptedTest(ValidatorTest):
    def test_accepted_rule_with_invalid_values_expect_error(self):
        rules = {
            "tos1": "accepted",
            "tos2": "accepted",
            "tos3": "accepted",
            "tos4": "accepted",
        }
        data = {"tos1": "off", "tos2": 2, "tos3": False, "tos4": "no"}
        expected = ACCEPTED_ERROR

        errors = self.validator.validate(data, rules)

        self.assertEqual(len(errors.items()), 4)
        for field, errs in errors.items():
            self.assertEqual(errs[0], expected.format(field=field))

    def test_accepted_rule_with_valid_values_expect_no_error(self):
        rules = {
            "tos1": "accepted",
            "tos2": "accepted",
            "tos3": "accepted",
            "tos4": "accepted",
        }
        data = {"tos1": "on", "tos2": 1, "tos3": True, "tos4": "yes"}
        expected = ACCEPTED_ERROR

        errors = self.validator.validate(data, rules)

        self.assertEqual(len(errors.items()), 0)
        for field, errs in errors.items():
            self.assertEqual(errs[0], expected.format(field=field))



from src.spotlight.errors import EMAIL_ERROR, MIN_STRING_ERROR
from .validator_test import ValidatorTest


class CombinationTest(ValidatorTest):
    def test_combo_with_invalid_email_and_min_expect_error(self):
        field = "field1"
        rules = {"field1": "filled|email|min:5|max:255"}
        data = {"field1": "oops"}

        errors = self.validator.validate(data, rules)
        errs = errors.get(field)

        self.assertEqual(errs[0], EMAIL_ERROR.format(field=field))
        self.assertEqual(errs[1], MIN_STRING_ERROR.format(field=field, min=5))

    def test_before_and_after_with_custom_format_expect_no_error(self):
        rules = {
            "start_time": "date_time:%H:%M:%S|after:12:00:00|before:end_time",
            "end_time": "date_time:%H:%M:%S",
        }
        data = {"start_time": "12:30:00", "end_time": "13:00:00"}
        expected = {}

        errors = self.validator.validate(data, rules)

        self.assertEqual(errors, expected)



from src.spotlight.errors import (
    MAX_STRING_ERROR,
    EMAIL_ERROR,
    FILLED_ERROR,
    IN_ERROR,
    REQUIRED_ERROR,
    MIN_STRING_ERROR,
)
from .validator_test import ValidatorTest


class NestedValidationTest(ValidatorTest):
    def setUp(self):
        self.validator.overwrite_messages = {}
        self.validator.overwrite_fields = {}
        self.validator.overwrite_values = {}

    def test_nested_validation_expect_error(self):
        rules = {"nested.test": "max:5|email", "non_nested": "max:1"}
        data = {"nested": {"test": "123456"}, "non_nested": "12"}
        expected = {
            "nested.test": [
                MAX_STRING_ERROR.format(field="nested.test", max=5),
                EMAIL_ERROR.format(field="nested.test"),
            ],
            "non_nested": [MAX_STRING_ERROR.format(field="non_nested", max=1)],
        }

        errors = self.validator.validate(data, rules)

        self.assertEqual(errors, expected)

    def test_double_nested_validation_expect_error(self):
        rules = {
            "nested.test": "max:5|email",
            "nested.double_nested.test2": "max:2",
            "non_nested": "max:1",
        }
        data = {
            "nested": {"test": "123456", "double_nested": {"test2": "12222"}},
            "non_nested": "12",
        }
        expected = {
            "nested.test": [
                MAX_STRING_ERROR.format(field="nested.test", max=5),
                EMAIL_ERROR.format(field="nested.test"),
            ],
            "nested.double_nested.test2": [
                MAX_STRING_ERROR.format(field="nested.double_nested.test2", max=2)
            ],
            "non_nested": [MAX_STRING_ERROR.format(field="non_nested", max=1)],
        }

        errors = self.validator.validate(data, rules)

        self.assertEqual(errors, expected)

    def test_double_nested_validation_custom_message_expect_error(self):
        rules = {
            "nested.test": "max:5|email",
            "nested.double_nested.test2": "max:2",
            "non_nested": "max:1",
        }
        data = {
            "nested": {"test": "123456", "double_nested": {"test2": "12222"}},
            "non_nested": "12",
        }
        new_message = "Hey! The {field} field has to be at least {max} chars!"
        expected = {
            "nested.test": [
                MAX_STRING_ERROR.format(field="nested.test", max=5),
                EMAIL_ERROR.format(field="nested.test"),
            ],
            "nested.double_nested.test2": [
                new_message.format(field="nested.double_nested.test2", max=2)
            ],
            "non_nested": [MAX_STRING_ERROR.format(field="non_nested", max=1)],
        }
        self.validator.overwrite_messages = {
            "nested.double_nested.test2.max": new_message
        }

        errors = self.validator.validate(data, rules)

        self.assertEqual(errors, expected)

    def test_double_nested_validation_custom_field_expect_error(self):
        rules = {
            "nested.test": "max:5|email",
            "nested.double_nested.test2": "max:2",
            "non_nested": "max:1",
        }
        data = {
            "nested": {"test": "123456", "double_nested": {"test2": "12222"}},
            "non_nested": "12",
        }
        expected = {
            "nested.double_nested.test2": [
                MAX_STRING_ERROR.format(field="custom", max=2)
            ],
            "nested.test": [
                MAX_STRING_ERROR.format(field="nested.test", max=5),
                EMAIL_ERROR.format(field="nested.test"),
            ],
            "non_nested": [MAX_STRING_ERROR.format(field="non_nested", max=1)],
        }
        self.validator.overwrite_fields = {"nested.double_nested.test2": "custom"}

        errors = self.validator.validate(data, rules)

        self.assertEqual(errors, expected)

    def test_double_nested_validation_with_new_values_expect_error(self):
        rules = {
            "nested.test": "max:5|email",
            "nested.double_nested.test2": "in:val1,val2,val3",
            "non_nested": "max:1",
        }
        data = {
            "nested": {"test": "123456", "double_nested": {"test2": "12222"}},
            "non_nested": "12",
        }
        new_values = "new1, new2, new3"
        expected = {
            "nested.double_nested.test2": [
                IN_ERROR.format(field="nested.double_nested.test2", values=new_values)
            ],
            "nested.test": [
                MAX_STRING_ERROR.format(field="nested.test", max=5),
                EMAIL_ERROR.format(field="nested.test"),
            ],
            "non_nested": [MAX_STRING_ERROR.format(field="non_nested", max=1)],
        }
        self.validator.overwrite_values = {
            "nested.double_nested.test2": {
                "val1": "new1",
                "val2": "new2",
                "val3": "new3",
            }
        }

        errors = self.validator.validate(data, rules)

        self.assertEqual(errors, expected)

    def test_nested_validation_with_nested_required_field_expect_error(self):
        rules = {"nested.test": "required"}
        data = {"nested": {}}
        expected = {"nested.test": [REQUIRED_ERROR.format(field="nested.test")]}

        errors = self.validator.validate(data, rules)

        self.assertEqual(errors, expected)

    def test_nested_validation_with_nested_field_expect_error(self):
        rules = {"nested.test": "max:5|email"}
        data = {"nested": {"test": "123456"}}
        expected = {
            "nested.test": [
                MAX_STRING_ERROR.format(field="nested.test", max=5),
                EMAIL_ERROR.format(field="nested.test"),
            ]
        }

        errors = self.validator.validate(data, rules)

        self.assertEqual(errors, expected)

    def test_double_nested_validation_with_flat_expect_error_list(self):
        rules = {
            "nested.test": "max:5|email",
            "nested.double_nested.test2": "in:val1,val2,val3",
            "non_nested": "max:1",
        }
        data = {
            "nested": {"test": "123456", "double_nested": {"test2": "12222"}},
            "non_nested": "12",
        }
        new_values = "new1, new2, new3"
        expected = [
            MAX_STRING_ERROR.format(field="nested.test", max=5),
            EMAIL_ERROR.format(field="nested.test"),
            IN_ERROR.format(field="nested.double_nested.test2", values=new_values),
            MAX_STRING_ERROR.format(field="non_nested", max=1),
        ]
        self.validator.overwrite_values = {
            "nested.double_nested.test2": {
                "val1": "new1",
                "val2": "new2",
                "val3": "new3",
            }
        }

        errors = self.validator.validate(data, rules, flat=True)

        self.assertEqual(errors, expected)

    def test_nested_validation_with_optional_field_and_empty_dict_expect_no_error(self):
        rules = {"nested.test": "email"}
        data = {"nested": {}}
        expected = {}

        errors = self.validator.validate(data, rules)

        self.assertEqual(errors, expected)

    def test_nested_validation_with_optional_field_and_input_none_expect_no_error(self):
        rules = {"nested.test": "email"}
        data = {"nested": None}
        expected = {}

        errors = self.validator.validate(data, rules)

        self.assertEqual(errors, expected)

    def test_nested_validation_with_optional_field_and_no_input_expect_no_error(self):
        rules = {"nested.test": "email"}
        data = {}
        expected = {}

        errors = self.validator.validate(data, rules)

        self.assertEqual(errors, expected)

    def test_nested_validation_with_filled_field_and_no_input_expect_error(self):
        rules = {"nested.test": "filled"}
        data = {"nested": {"test": None}}
        expected = {"nested.test": [FILLED_ERROR.format(field="nested.test")]}

        errors = self.validator.validate(data, rules)

        self.assertEqual(errors, expected)

    def test_nested_validation_with_required_field_expect_no_error(self):
        rules = {"nested.test": "required"}
        data = {"nested": {"test": "test"}}
        expected = {}

        errors = self.validator.validate(data, rules)

        self.assertEqual(errors, expected)

    def test_nested_validation_with_list(self):
        rules = {
            "nested.list.*.test": "required|max:2",
            "nested.list.*.test2": "min:2",
            "non_nested": "max:2",
        }
        data = {
            "nested": {
                "list": [
                    {"test": "12", "test2": "1"},
                    {"test2": "1"},
                    {"test": "1234", "test2": "1"},
                ]
            },
            "non_nested": "test",
        }
        expected = {
            "nested.list.1.test": [REQUIRED_ERROR.format(field="nested.list.1.test")],
            "nested.list.2.test": [
                MAX_STRING_ERROR.format(field="nested.list.2.test", max=2)
            ],
            "nested.list.0.test2": [
                MIN_STRING_ERROR.format(field="nested.list.0.test2", min=2)
            ],
            "nested.list.1.test2": [
                MIN_STRING_ERROR.format(field="nested.list.1.test2", min=2)
            ],
            "nested.list.2.test2": [
                MIN_STRING_ERROR.format(field="nested.list.2.test2", min=2)
            ],
            "non_nested": [MAX_STRING_ERROR.format(field="non_nested", max=2)],
        }

        errors = self.validator.validate(data, rules)

        self.assertEqual(errors, expected)



from src.spotlight.errors import ALPHA_NUM_SPACE_ERROR
from .validator_test import ValidatorTest


class AlphaNumSpaceTest(ValidatorTest):
    def test_alpha_num_space_rule_with_invalid_values_expect_errors(self):
        rules = {
            "field1": "alpha_num_space",
            "field2": "alpha_num_space",
            "field3": "alpha_num_space",
            "field4": "alpha_num_space",
        }
        data = {
            "field1": "!@#$%^&*",
            "field2": "alpha-123",
            "field3": "alpha_123",
            "field4": 1,
        }

        errors = self.validator.validate(data, rules)

        self.assertEqual(len(errors.items()), 4)
        for field, errs in errors.items():
            expected = ALPHA_NUM_SPACE_ERROR.format(field=field)
            self.assertEqual(errs[0], expected)

    def test_alpha_num_space_rule_with_valid_values_expect_no_errors(self):
        rules = {
            "field1": "alpha_num_space",
            "field2": "alpha_num_space",
            "field3": "alpha_num_space",
        }
        data = {"field1": "test", "field2": "1", "field3": "alpha 123"}
        expected = None

        errors = self.validator.validate(data, rules)

        self.assertEqual(len(errors.items()), 0)
        for field, errs in errors.items():
            self.assertEqual(errs, expected)

    def test_valid_alpha_num_space_with_invalid_values_expect_false(self):
        values = ["!@#$%^&*", "alpha-123", "alpha_123", 1, [], {}, True, False]

        for value in values:
            actual = self.validator.valid_alpha_num_space(value)
            self.assertEqual(actual, False)

    def test_valid_alpha_num_space_with_valid_values_expect_true(self):
        values = ["test", "1", "alpha 123"]

        for value in values:
            actual = self.validator.valid_alpha_num_space(value)
            self.assertEqual(actual, True)



from src.spotlight.errors import REQUIRED_IF_ERROR
from .validator_test import ValidatorTest


class RequiredIfTest(ValidatorTest):
    def setUp(self):
        self.other_field = "test1"
        self.field = "test2"
        self.value = "test"
        self.required_if_error = REQUIRED_IF_ERROR.format(
            field=self.field, other=self.other_field, value=self.value
        )

    def test_required_if_rule_with_other_field_string_expect_error(self):
        data = {"test1": "test"}
        rules = {"test2": "required_if:test1,test"}
        expected = self.required_if_error

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_required_if_rule_with_field_present_expect_no_error(self):
        data = {"test1": "test", "test2": "world"}
        rules = {"test2": "required_if:test1,test"}
        expected = None

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)

    def test_required_if_rule_with_wrong_value_no_error(self):
        data = {"test1": "test2"}
        rules = {"test2": "required_if:test1,test"}
        expected = None

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)

    def test_required_if_rule_with_boolean_true_expect_error(self):
        data = {"test1": True}
        rules = {"test2": "required_if:test1,True"}
        expected = REQUIRED_IF_ERROR.format(
            field=self.field, other=self.other_field, value=True
        )

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_required_if_rule_with_boolean_false_expect_error(self):
        data = {"test1": False}
        rules = {"test2": "required_if:test1,False"}
        expected = REQUIRED_IF_ERROR.format(
            field=self.field, other=self.other_field, value=False
        )

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_required_if_rule_with_boolean_true_expect_no_error(self):
        data = {"test1": True, "test2": "I am required"}
        rules = {"test2": "required_if:test1,true"}
        expected = None

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)

    def test_required_if_rule_with_boolean_false_expect_no_error(self):
        data = {"test1": False, "test2": "I am required"}
        rules = {"test2": "required_if:test1,false"}
        expected = None

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)

    def test_required_if_rule_with_integer_expect_no_error(self):
        data = {"test1": 1, "test2": "I am required"}
        rules = {"test2": "required_if:test1,1"}
        expected = None

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)

    def test_required_if_rule_with_integer_expect_error(self):
        data = {"test1": 1}
        rules = {"test2": "required_if:test1,1"}
        expected = REQUIRED_IF_ERROR.format(
            field=self.field, other=self.other_field, value=1
        )

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)



from src.spotlight.errors import EMAIL_ERROR
from .validator_test import ValidatorTest


class EmailTest(ValidatorTest):
    def setUp(self):
        self.field = "email"
        self.email_error = EMAIL_ERROR.format(field=self.field)

    def test_email_rule_with_invalid_email_expect_error(self):
        rules = {"email": "email"}
        data = {"email": "this.is.not.a.valid.email"}
        expected = self.email_error

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_email_rule_with_valid_email_expect_no_error(self):
        rules = {"email": "email"}
        data = {"email": "john.doe@example.com"}
        expected = None

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)

    def test_email_rule_with_empty_string_expect_error(self):
        rules = {"email": "email"}
        data = {"email": ""}
        expected = self.email_error

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs[0], expected)

    def test_email_rule_with_no_input_expect_no_error(self):
        rules = {"email": "email"}
        data = {}
        expected = None

        errors = self.validator.validate(data, rules)
        errs = errors.get(self.field)

        self.assertEqual(errs, expected)

    def test_valid_email_with_invalid_values_expect_false(self):
        invalid_emails = [1, True, False, "test", "test@", "@"]

        for invalid_email in invalid_emails:
            actual = self.validator.valid_email(invalid_email)
            self.assertEqual(actual, False)

    def test_valid_email_with_valid_values_expect_true(self):
        valid_emails = ["a@a", "test@example", "test@example.com"]

        for valid_email in valid_emails:
            actual = self.validator.valid_email(valid_email)
            self.assertEqual(actual, True)



from typing import Any, List

from src.spotlight.rules import Rule
from .validator_test import ValidatorTest


class ExactlyFiveCharsRule(Rule):
    """Exactly 5 characters"""

    name = "five_chars"

    def passes(self, field: str, value: Any, parameters: List[str], validator) -> bool:
        self.message_fields = dict(field=field, name="you.can.replace.this")

        return len(value) == 5

    @property
    def message(self) -> str:
        return "The {field} field has to be exactly 5 chars long {name}!"


class UppercaseRule(Rule):
    """Uppercase"""

    name = "uppercase"

    def passes(self, field: str, value: Any, parameters: List[str], validator) -> bool:
        self.message_fields = dict(field=field)

        return value.upper() == value

    @property
    def message(self) -> str:
        return "The {field} field must be uppercase."


class CustomRuleTest(ValidatorTest):
    def setUp(self):
        self.validator.fields = {}

    def test_custom_rule_uppercase_with_non_uppercase_string_expect_error(self):
        field = "test"
        rules = {"test": "uppercase"}
        data = {"test": "test"}
        expected = "The test field must be uppercase."

        rule = UppercaseRule()
        self.validator.register_rule(rule)
        errors = self.validator.validate(data, rules)
        errs = errors.get(field)

        self.assertEqual(errs[0], expected)

    def test_custom_rule_five_chars_with_invalid_string_expect_error(self):
        field = "test"
        rules = {"test": "five_chars"}
        data = {"test": "test"}
        rule = ExactlyFiveCharsRule()
        self.validator.overwrite_fields = {
            "test": "test",
            "you.can.replace.this": "custom",
        }
        expected = rule.message.format(field=field, name="custom")

        self.validator.register_rule(rule)
        errors = self.validator.validate(data, rules)
        errs = errors.get(field)

        self.assertEqual(errs[0], expected)



from src.spotlight.errors import URL_ERROR
from .validator_test import ValidatorTest


class UrlTest(ValidatorTest):
    def test_url_rule_with_invalid_urls_expect_errors(self):
        rules = {"url1": "url", "url2": "url", "url3": "url"}
        data = {
            "url1": "this.is.not.a.valid.url",
            "url2": "http",
            "url3": "www.google.com",
        }

        errors = self.validator.validate(data, rules)

        self.assertEqual(len(errors.items()), 3)
        for field, errs in errors.items():
            expected = URL_ERROR.format(field=field)
            self.assertEqual(errs[0], expected)

    def test_url_rule_with_valid_urls_expect_no_errors(self):
        rules = {
            "url1": "url",
            "url2": "url",
            "url3": "url",
            "url4": "url",
            "url5": "url",
            "url6": "url",
            "url7": "url",
        }
        data = {
            "url1": "http://google.com",
            "url2": "https://google.com",
            "url3": "http://www.google.com",
            "url4": "http://localhost",
            "url5": "http://localhost:8080",
            "url6": "http://localhost:8080/test/",
            "url7": "http://test.dev/test/",
        }
        expected = None

        errors = self.validator.validate(data, rules)

        self.assertEqual(len(errors.items()), 0)
        for field, errs in errors.items():
            self.assertEqual(errs, expected)

    def test_valid_url_with_boolean_true_expect_false(self):
        valid_url = self.validator.valid_url(True)

        self.assertEqual(valid_url, False)

    def test_valid_url_with_valid_url_expect_true(self):
        valid_url = self.validator.valid_url("https://google.com")

        self.assertEqual(valid_url, True)
